"""DXT decompression: DXT1/DXT3/DXT5 block decode to RGB/RGBA."""

from __future__ import annotations


def dxt1_to_rgb(dxt1_data: bytes | bytearray, width: int, height: int) -> bytearray:
    def unpack_565(color):
        r = (color >> 11) & 0x1F
        g = (color >> 5) & 0x3F
        b = color & 0x1F
        return (r << 3) | (r >> 2), (g << 2) | (g >> 4), (b << 3) | (b >> 2)

    def interpolate_colors(c0, c1, t):
        return tuple((1 - t) * a + t * b for a, b in zip(c0, c1))

    rgb_data = bytearray(width * height * 3)
    block_count_x = (width + 3) // 4
    block_count_y = (height + 3) // 4
    data_length = len(dxt1_data)

    for block_y in range(block_count_y):
        for block_x in range(block_count_x):
            block_offset = (block_y * block_count_x + block_x) * 8

            # Check if we have enough data for this block (8 bytes for DXT1)
            if block_offset >= data_length or block_offset + 8 > data_length:
                # Fill remaining pixels with black
                for y in range(4):
                    for x in range(4):
                        pixel_x = block_x * 4 + x
                        pixel_y = block_y * 4 + y
                        if pixel_x < width and pixel_y < height:
                            pixel_offset = (pixel_y * width + pixel_x) * 3
                            rgb_data[pixel_offset : pixel_offset + 3] = [0, 0, 0]
                continue

            color0 = int.from_bytes(dxt1_data[block_offset : block_offset + 2], "little")
            color1 = int.from_bytes(dxt1_data[block_offset + 2 : block_offset + 4], "little")
            color_indices = int.from_bytes(dxt1_data[block_offset + 4 : block_offset + 8], "little")

            c0 = unpack_565(color0)
            c1 = unpack_565(color1)

            color_table = [
                c0,
                c1,
                interpolate_colors(c0, c1, 2 / 3) if color0 > color1 else interpolate_colors(c0, c1, 1 / 2),
                interpolate_colors(c0, c1, 1 / 3) if color0 > color1 else (0, 0, 0),
            ]

            for y in range(4):
                for x in range(4):
                    pixel_x = block_x * 4 + x
                    pixel_y = block_y * 4 + y
                    if pixel_x < width and pixel_y < height:
                        color_index = (color_indices >> (2 * (y * 4 + x))) & 0x3
                        color = color_table[color_index]
                        pixel_offset = (pixel_y * width + pixel_x) * 3
                        rgb_data[pixel_offset : pixel_offset + 3] = [int(c) for c in color]

    return rgb_data


def dxt3_to_rgba(dxt3_data: bytes | bytearray, width: int, height: int) -> bytearray:
    def unpack_565(color: int) -> tuple[int, int, int]:
        r: int = (color >> 11) & 0x1F
        g: int = (color >> 5) & 0x3F
        b: int = color & 0x1F
        return (r << 3) | (r >> 2), (g << 2) | (g >> 4), (b << 3) | (b >> 2)

    def lerp(a: int, b: int, t: float) -> int:
        return int(a + (b - a) * t)

    rgba = bytearray(width * height * 4)
    block_count_x = (width + 3) // 4
    block_count_y = (height + 3) // 4
    data_length = len(dxt3_data)

    for block_y in range(block_count_y):
        for block_x in range(block_count_x):
            block_offset = (block_y * block_count_x + block_x) * 16

            # Check if we have enough data for this block (16 bytes for DXT3)
            if block_offset >= data_length or block_offset + 16 > data_length:
                # Fill remaining pixels with transparent black
                for y in range(4):
                    for x in range(4):
                        pixel_x = block_x * 4 + x
                        pixel_y = block_y * 4 + y
                        if pixel_x < width and pixel_y < height:
                            rgba_index = ((block_y * 4 + y) * width + block_x * 4 + x) * 4
                            rgba[rgba_index : rgba_index + 4] = [0, 0, 0, 0]
                continue

            alpha_values = dxt3_data[block_offset : block_offset + 8]
            color0 = int.from_bytes(dxt3_data[block_offset + 8 : block_offset + 10], "little")
            color1 = int.from_bytes(dxt3_data[block_offset + 10 : block_offset + 12], "little")
            color_indices = int.from_bytes(dxt3_data[block_offset + 12 : block_offset + 16], "little")

            r0, g0, b0 = unpack_565(color0)
            r1, g1, b1 = unpack_565(color1)

            color_table: list[tuple[int, int, int]] = [
                (r0, g0, b0),
                (r1, g1, b1),
                (lerp(r0, r1, 2 / 3), lerp(g0, g1, 2 / 3), lerp(b0, b1, 2 / 3)),
                (lerp(r0, r1, 1 / 3), lerp(g0, g1, 1 / 3), lerp(b0, b1, 1 / 3)),
            ]

            for y in range(4):
                for x in range(4):
                    pixel_index = x + y * 4
                    rgba_index = ((block_y * 4 + y) * width + block_x * 4 + x) * 4

                    alpha = (alpha_values[pixel_index // 2] >> (4 * (pixel_index % 2))) & 0xF
                    alpha = (alpha << 4) | alpha

                    color_index = (color_indices >> (2 * (15 - pixel_index))) & 0x3
                    color = color_table[color_index]

                    rgba[rgba_index] = color[0]
                    rgba[rgba_index + 1] = color[1]
                    rgba[rgba_index + 2] = color[2]
                    rgba[rgba_index + 3] = alpha

    return rgba


def dxt5_to_rgba(
    dxt5_data: bytes | bytearray,
    width: int,
    height: int,
) -> bytearray:
    def unpack_565(color: int) -> tuple[int, int, int]:
        r: int = (color >> 11) & 0x1F
        g: int = (color >> 5) & 0x3F
        b: int = color & 0x1F
        return (r << 3 | r >> 2, g << 2 | g >> 4, b << 3 | b >> 2)

    def interpolate_color(c0: tuple[int, int, int], c1: tuple[int, int, int], t: float) -> tuple[int, int, int]:
        return (int((1 - t) * c0[0] + t * c1[0]), int((1 - t) * c0[1] + t * c1[1]), int((1 - t) * c0[2] + t * c1[2]))

    def interpolate_alpha(a0: int, a1: int, t: float) -> int:
        return int((1 - t) * a0 + t * a1)

    rgba: bytearray = bytearray(width * height * 4)
    block_count_x: int = (width + 3) // 4
    block_count_y: int = (height + 3) // 4
    data_length: int = len(dxt5_data)

    for block_y in range(block_count_y):
        for block_x in range(block_count_x):
            block_offset: int = (block_y * block_count_x + block_x) * 16

            # Check if we have enough data for this block (16 bytes for DXT5)
            if block_offset >= data_length or block_offset + 16 > data_length:
                # Fill remaining pixels with transparent black
                for y in range(4):
                    for x in range(4):
                        pixel_x = block_x * 4 + x
                        pixel_y = block_y * 4 + y
                        if pixel_x < width and pixel_y < height:
                            fill_offset: int = (pixel_y * width + pixel_x) * 4
                            rgba[fill_offset : fill_offset + 4] = (0, 0, 0, 0)
                continue

            # Alpha
            alpha0: int = dxt5_data[block_offset]
            alpha1: int = dxt5_data[block_offset + 1]
            alpha_bits: int = int.from_bytes(dxt5_data[block_offset + 2 : block_offset + 8], "little")

            # Color
            color0: int = int.from_bytes(dxt5_data[block_offset + 8 : block_offset + 10], "little")
            color1: int = int.from_bytes(dxt5_data[block_offset + 10 : block_offset + 12], "little")
            color_bits: int = int.from_bytes(dxt5_data[block_offset + 12 : block_offset + 16], "little")

            c0: tuple[int, int, int] = unpack_565(color0)
            c1: tuple[int, int, int] = unpack_565(color1)

            for y in range(4):
                for x in range(4):
                    pixel_offset: int = ((block_y * 4 + y) * width + (block_x * 4 + x)) * 4

                    if block_x * 4 + x >= width or block_y * 4 + y >= height:
                        continue

                    color_index: int = (color_bits >> (2 * (4 * y + x))) & 0x3
                    alpha_index: int = (alpha_bits >> (3 * (4 * y + x))) & 0x7

                    if color_index == 0:
                        color: tuple[int, int, int] = c0
                    elif color_index == 1:
                        color = c1
                    elif color_index == 2:  # noqa: PLR2004
                        color = interpolate_color(c0, c1, 1 / 3)
                    else:
                        color = interpolate_color(c0, c1, 2 / 3)

                    if alpha_index == 0:
                        alpha = alpha0
                    elif alpha_index == 1:
                        alpha = alpha1
                    elif alpha0 > alpha1:
                        alpha = interpolate_alpha(alpha0, alpha1, alpha_index / 7)
                    elif alpha_index == 6:  # noqa: PLR2004
                        alpha = 0
                    elif alpha_index == 7:  # noqa: PLR2004
                        alpha = 255
                    else:
                        alpha = interpolate_alpha(alpha0, alpha1, (alpha_index - 1) / 5)

                    rgba[pixel_offset : pixel_offset + 4] = (*color, alpha)

    return rgba



"""This file was generated by pyuic6 from the .ui file provided from Qt's physical source code.

This file should be considered a template and should be modified to fit the needs of the project.

"""

from __future__ import annotations

from typing import TYPE_CHECKING, Callable

from qtpy import QtCore
from qtpy.QtWidgets import (
    QApplication,
    QComboBox,
    QDialogButtonBox,
    QFrame,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QSizePolicy,
    QSplitter,
    QStackedWidget,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from utility.gui.qt.adapters.filesystem.qfiledialog.private.qfiledialog_p import (
    QFileDialogComboBox,
    QFileDialogLineEdit,
    QFileDialogListView,
    QFileDialogTreeView,
)
from utility.gui.qt.adapters.filesystem.qfiledialog.private.qsidebar_p import QSidebar

if TYPE_CHECKING:
    from qtpy.QtWidgets import QFileDialog


class Ui_QFileDialog:  # noqa: N801
    def setupUi(self, QFileDialog: QFileDialog):  # noqa: N803, PLR0915
        QFileDialog.setObjectName("QFileDialog")
        QFileDialog.resize(521, 316)
        QFileDialog.setSizeGripEnabled(True)

        # Main layout
        self.gridlayout: QGridLayout = QGridLayout(QFileDialog)
        self.gridlayout.setObjectName("gridlayout")

        # Look In label
        self.lookInLabel: QLabel = QLabel(parent=QFileDialog)
        self.lookInLabel.setObjectName("lookInLabel")
        self.gridlayout.addWidget(self.lookInLabel, 0, 0, 1, 1)

        # Toolbar layout
        self.hboxlayout: QHBoxLayout = QHBoxLayout()
        self.hboxlayout.setObjectName("hboxlayout")

        # Look In combo box
        self.lookInCombo: QFileDialogComboBox = QFileDialogComboBox(parent=QFileDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lookInCombo.sizePolicy().hasHeightForWidth())
        self.lookInCombo.setSizePolicy(sizePolicy)
        self.lookInCombo.setMinimumSize(QtCore.QSize(50, 0))
        self.lookInCombo.setObjectName("lookInCombo")
        self.hboxlayout.addWidget(self.lookInCombo)

        # Toolbar buttons
        self.backButton: QToolButton = QToolButton(parent=QFileDialog)
        self.backButton.setObjectName("backButton")
        self.hboxlayout.addWidget(self.backButton)

        self.forwardButton: QToolButton = QToolButton(parent=QFileDialog)
        self.forwardButton.setObjectName("forwardButton")
        self.hboxlayout.addWidget(self.forwardButton)

        self.toParentButton: QToolButton = QToolButton(parent=QFileDialog)
        self.toParentButton.setObjectName("toParentButton")
        self.hboxlayout.addWidget(self.toParentButton)

        self.newFolderButton: QToolButton = QToolButton(parent=QFileDialog)
        self.newFolderButton.setObjectName("newFolderButton")
        self.hboxlayout.addWidget(self.newFolderButton)

        self.listModeButton: QToolButton = QToolButton(parent=QFileDialog)
        self.listModeButton.setObjectName("listModeButton")
        self.hboxlayout.addWidget(self.listModeButton)

        self.detailModeButton: QToolButton = QToolButton(parent=QFileDialog)
        self.detailModeButton.setObjectName("detailModeButton")
        self.hboxlayout.addWidget(self.detailModeButton)

        self.gridlayout.addLayout(self.hboxlayout, 0, 1, 1, 2)

        # Splitter
        self.splitter: QSplitter = QSplitter(parent=QFileDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.splitter.sizePolicy().hasHeightForWidth())
        self.splitter.setSizePolicy(sizePolicy)
        self.splitter.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.splitter.setChildrenCollapsible(False)
        self.splitter.setObjectName("splitter")

        # Sidebar
        self.sidebar: QSidebar = QSidebar(parent=self.splitter)
        self.sidebar.setObjectName("sidebar")

        # Main frame
        self.frame: QFrame = QFrame(parent=self.splitter)
        self.frame.setFrameShape(QFrame.Shape.NoFrame)
        self.frame.setFrameShadow(QFrame.Shadow.Raised)
        self.frame.setObjectName("frame")

        self.vboxlayout: QVBoxLayout = QVBoxLayout(self.frame)
        self.vboxlayout.setContentsMargins(0, 0, 0, 0)
        self.vboxlayout.setSpacing(0)
        self.vboxlayout.setObjectName("vboxlayout")

        # Stacked widget for list and tree views
        self.stackedWidget: QStackedWidget = QStackedWidget(parent=self.frame)
        self.stackedWidget.setObjectName("stackedWidget")

        # List view page
        self.page: QWidget = QWidget()
        self.page.setObjectName("page")
        self.vboxlayout1: QVBoxLayout = QVBoxLayout(self.page)
        self.vboxlayout1.setContentsMargins(0, 0, 0, 0)
        self.vboxlayout1.setSpacing(0)
        self.vboxlayout1.setObjectName("vboxlayout1")
        self.listView: QFileDialogListView = QFileDialogListView(parent=self.page)
        self.listView.setObjectName("listView")
        self.vboxlayout1.addWidget(self.listView)
        self.stackedWidget.addWidget(self.page)

        # Tree view page
        self.page_2: QWidget = QWidget()
        self.page_2.setObjectName("page_2")
        self.vboxlayout2: QVBoxLayout = QVBoxLayout(self.page_2)
        self.vboxlayout2.setContentsMargins(0, 0, 0, 0)
        self.vboxlayout2.setSpacing(0)
        self.vboxlayout2.setObjectName("vboxlayout2")
        self.treeView: QFileDialogTreeView = QFileDialogTreeView(parent=self.page_2)
        self.treeView.setObjectName("treeView")
        self.vboxlayout2.addWidget(self.treeView)
        self.stackedWidget.addWidget(self.page_2)

        self.vboxlayout.addWidget(self.stackedWidget)
        self.gridlayout.addWidget(self.splitter, 1, 0, 1, 3)

        # File name label and edit
        self.fileNameLabel: QLabel = QLabel(parent=QFileDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fileNameLabel.sizePolicy().hasHeightForWidth())
        self.fileNameLabel.setSizePolicy(sizePolicy)
        self.fileNameLabel.setMinimumSize(QtCore.QSize(0, 0))
        self.fileNameLabel.setObjectName("fileNameLabel")
        self.gridlayout.addWidget(self.fileNameLabel, 2, 0, 1, 1)

        self.fileNameEdit: QFileDialogLineEdit = QFileDialogLineEdit(parent=QFileDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fileNameEdit.sizePolicy().hasHeightForWidth())
        self.fileNameEdit.setSizePolicy(sizePolicy)
        self.fileNameEdit.setObjectName("fileNameEdit")
        self.gridlayout.addWidget(self.fileNameEdit, 2, 1, 1, 1)

        # Button box
        self.buttonBox: QDialogButtonBox = QDialogButtonBox(parent=QFileDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Vertical)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridlayout.addWidget(self.buttonBox, 2, 2, 2, 1)

        # File type label and combo
        self.fileTypeLabel: QLabel = QLabel(parent=QFileDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fileTypeLabel.sizePolicy().hasHeightForWidth())
        self.fileTypeLabel.setSizePolicy(sizePolicy)
        self.fileTypeLabel.setObjectName("fileTypeLabel")
        self.gridlayout.addWidget(self.fileTypeLabel, 3, 0, 1, 1)

        self.fileTypeCombo: QComboBox = QComboBox(parent=QFileDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fileTypeCombo.sizePolicy().hasHeightForWidth())
        self.fileTypeCombo.setSizePolicy(sizePolicy)
        self.fileTypeCombo.setObjectName("fileTypeCombo")
        self.gridlayout.addWidget(self.fileTypeCombo, 3, 1, 1, 1)

        self.retranslateUi(QFileDialog)
        self.stackedWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(QFileDialog)
        QFileDialog.setTabOrder(self.lookInCombo, self.backButton)
        QFileDialog.setTabOrder(self.backButton, self.forwardButton)
        QFileDialog.setTabOrder(self.forwardButton, self.toParentButton)
        QFileDialog.setTabOrder(self.toParentButton, self.newFolderButton)
        QFileDialog.setTabOrder(self.newFolderButton, self.listModeButton)
        QFileDialog.setTabOrder(self.listModeButton, self.detailModeButton)
        QFileDialog.setTabOrder(self.detailModeButton, self.sidebar)
        QFileDialog.setTabOrder(self.sidebar, self.treeView)
        QFileDialog.setTabOrder(self.treeView, self.listView)
        QFileDialog.setTabOrder(self.listView, self.fileNameEdit)
        QFileDialog.setTabOrder(self.fileNameEdit, self.buttonBox)
        QFileDialog.setTabOrder(self.buttonBox, self.fileTypeCombo)

    def retranslateUi(self, QFileDialog: QFileDialog):  # noqa: N803
        _translate: Callable[[str, str], str] = QApplication.instance().translate  # pyright: ignore[reportOptionalMemberAccess]
        self.lookInLabel.setText(_translate("QFileDialog", "Look in:"))

        self.backButton.setToolTip(_translate("QFileDialog", "Back"))
        self.backButton.setAccessibleName(_translate("QFileDialog", "Back"))
        self.backButton.setAccessibleDescription(_translate("QFileDialog", "Go back"))
        self.backButton.setShortcut(_translate("QFileDialog", "Alt+Left"))

        self.forwardButton.setToolTip(_translate("QFileDialog", "Forward"))
        self.forwardButton.setAccessibleName(_translate("QFileDialog", "Forward"))
        self.forwardButton.setAccessibleDescription(_translate("QFileDialog", "Go forward"))
        self.forwardButton.setShortcut(_translate("QFileDialog", "Alt+Right"))

        self.toParentButton.setToolTip(_translate("QFileDialog", "Parent Directory"))
        self.toParentButton.setAccessibleName(_translate("QFileDialog", "Parent Directory"))
        self.toParentButton.setAccessibleDescription(_translate("QFileDialog", "Go to the parent directory"))
        self.toParentButton.setShortcut(_translate("QFileDialog", "Alt+Up"))

        self.newFolderButton.setToolTip(_translate("QFileDialog", "Create New Folder"))
        self.newFolderButton.setAccessibleName(_translate("QFileDialog", "Create New Folder"))
        self.newFolderButton.setAccessibleDescription(_translate("QFileDialog", "Create a New Folder"))

        self.listModeButton.setToolTip(_translate("QFileDialog", "List View"))
        self.listModeButton.setAccessibleName(_translate("QFileDialog", "List View"))
        self.listModeButton.setAccessibleDescription(_translate("QFileDialog", "Change to list view mode"))

        self.detailModeButton.setToolTip(_translate("QFileDialog", "Detail View"))
        self.detailModeButton.setAccessibleName(_translate("QFileDialog", "Detail View"))
        self.detailModeButton.setAccessibleDescription(_translate("QFileDialog", "Change to detail view mode"))

        self.sidebar.setAccessibleName(_translate("QFileDialog", "Sidebar"))
        self.sidebar.setAccessibleDescription(_translate("QFileDialog", "List of places and bookmarks"))

        self.listView.setAccessibleName(_translate("QFileDialog", "Files"))
        self.treeView.setAccessibleName(_translate("QFileDialog", "Files"))
        self.fileTypeLabel.setText(_translate("QFileDialog", "Files of type:"))



"""UTP (placeable) generic: GFF-based placeable definitions and inventory."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pykotor.common.language import LocalizedString
from pykotor.common.misc import Game, InventoryItem, ResRef
from pykotor.resource.formats.gff import GFF, GFFContent, GFFList, read_gff, write_gff
from pykotor.resource.formats.gff.gff_auto import bytes_gff
from pykotor.resource.formats.gff.gff_data import GFFStruct
from pykotor.resource.type import ResourceType

if TYPE_CHECKING:
    from pykotor.resource.formats.gff.gff_data import GFFStruct
    from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES


class UTP:
    """Stores placeable data.

    UTP files are GFF-based format files that store placeable object definitions including
    lock/unlock mechanics, HP, inventory, scripts, and appearance.

    References:
    ----------
        Based on unified K1 (swkotor.exe) and TSL (swkotor2.exe) UTP implementation.
        Addresses: (K1: swkotor.exe, TSL: swkotor2.exe). TSL addresses: resolve in REVA when
        PyKotorGhidraProject.gpr is open (project may be locked by another process).

        - CSWSPlaceable::LoadPlaceable (main UTP GFF parser)
            K1: 0x00585670, TSL: 0x006a1680 (LoadPlaceableFromGFF, legacy PC)
            Loads all placeable fields from GFF structure.
            Signature: LoadPlaceable(CSWSPlaceable* this, CResGFF* param_1, CResStruct* param_2, int param_3).
            Called from LoadPlaceables and LoadFromTemplate.

        - CSWSPlaceable::SavePlaceable (UTP GFF writer)
            K1: 0x00586a70, TSL: TODO
            Writes all placeable fields to GFF structure.

        - LoadPlaceables / LoadFromTemplate (callers)
            K1: LoadPlaceables 0x0050a7b0, LoadFromTemplate 0x00587a70; TSL: TODO

        - GFF field label string references (K1; TSL at different addresses, TODO):
            K1: "Appearance" 0x00746efc, "HasInventory" 0x007496e0, "Lockable" 0x007496f8,
            "KeyName" 0x0074979c, "OpenLockDC" 0x007497a4, "CloseLockDC" 0x007496c8.

        GFF Field Structure (from LoadPlaceable analysis):
            - Root struct fields:
                - "Tag" (CExoString) - Placeable tag identifier
                - "TemplateResRef" (CResRef) - Template resource reference
                - "LocName" (CExoLocString) - Localized placeable name
                - "AutoRemoveKey" (BYTE) - Whether key is auto-removed after use
                - "Faction" (DWORD) - Faction identifier
                - "Invulnerable" (BYTE) - Whether placeable is invulnerable
                - "Plot" (BYTE) - Whether placeable is plot-critical
                - "Min1HP" (BYTE) - Whether placeable has minimum 1 HP
                - "PartyInteract" (BYTE) - Whether party can interact
                - "OpenLockDC" (BYTE) - Open lock difficulty class
                - "KeyName" (CExoString) - Key name/ResRef
                - "TrapDisarmable" (BYTE) - Whether trap is disarmable
                - "TrapDetectable" (BYTE) - Whether trap is detectable
                - "DisarmDC" (BYTE) - Disarm difficulty class
                - "TrapDetectDC" (BYTE) - Trap detection difficulty class
                - "TrapFlag" (BYTE) - Trap flag
                - "TrapOneShot" (BYTE) - Whether trap is one-shot
                - "TrapType" (BYTE) - Trap type identifier
                - "Useable" (BYTE) - Whether placeable is usable
                - "Static" (BYTE) - Whether placeable is static
                - "Appearance" (DWORD) - Appearance type identifier
                - "HP" (SHORT) - Hit points
                - "CurrentHP" (SHORT) - Current hit points
                - "Hardness" (BYTE) - Hardness value
                - "Fort" (BYTE) - Fortitude save
                - "Will" (BYTE) - Will save
                - "Ref" (BYTE) - Reflex save
                - "Lockable" (BYTE) - Whether placeable is lockable
                - "Locked" (BYTE) - Whether placeable is locked
                - "HasInventory" (BYTE) - Whether placeable has inventory
                - "KeyRequired" (BYTE) - Whether key is required
                - "CloseLockDC" (BYTE) - Close lock difficulty class
                - "PortraitId" (WORD) - Portrait ID (0xffff = use Portrait ResRef)
                - "Portrait" (CResRef) - Portrait resource reference (if PortraitId == 0xffff)
                - "Conversation" (CResRef) - Conversation dialog ResRef
                - "BodyBag" (BYTE) - Body bag type
                - "DieWhenEmpty" (BYTE) - Whether placeable dies when inventory empty
                - "GroundPile" (BYTE) - Ground pile flag
                - "LightState" (BYTE) - Light state (on/off)
                - "Description" (CExoLocString) - Placeable description
                - "ItemList" (GFFList) - List of inventory items (if HasInventory)

        Note: UTP files are GFF format files with specific structure definitions (GFFContent.UTP)

    Derivations and Other Implementations:
    ----------
        https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Resources/KotorUTP/UTP.cs:12-75 (UTP class definition)

    Attributes:
    ----------
        resref: "TemplateResRef" field. The resource reference for this placeable template.
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/UTP.cs:64 (TemplateResRef property)

        tag: "Tag" field. Tag identifier for this placeable.
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/UTP.cs:63 (Tag property)

        name: "LocName" field. Localized name of the placeable.
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/UTP.cs:35 (LocName property)

        appearance_id: "Appearance" field. Placeable appearance type identifier.
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/UTP.cs:15 (Appearance property)

        has_inventory: "HasInventory" field. Whether placeable has an inventory.
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/UTP.cs:27 (HasInventory property)

        inventory: List of InventoryItem objects in this placeable's inventory.
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/UTP.cs:74 (Inventory property)

        not_blastable: "NotBlastable" field. Whether placeable cannot be blasted. KotOR 2 Only.
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/UTP.cs:37 (NotBlastable property)
            Reference: NorthernLights/AuroraUTP.cs:67 (NotBlastable field)

        unlock_diff: "OpenLockDiff" field. Unlock difficulty modifier. KotOR 2 Only.
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/UTP.cs:55 (OpenLockDiff property)
            Reference: NorthernLights/AuroraUTP.cs:68 (OpenLockDiff field)

        unlock_diff_mod: "OpenLockDiffMod" field. Additional unlock difficulty modifier. KotOR 2 Only.
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/UTP.cs:56 (OpenLockDiffMod property as sbyte)
            Reference: NorthernLights/AuroraUTP.cs:69 (OpenLockDiffMod field as Char)
            Note: Type discrepancy - reone uses char/int, Kotor.NET uses sbyte, PyKotor uses int

        on_open_failed: "OnFailToOpen" field. Script to run when placeable fails to open. KotOR 2 Only.
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/UTP.cs:43 (OnFailToOpen property)

        lock_dc: "CloseLockDC" field. Difficulty class to lock placeable. KotOR 2 Only.
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/UTP.cs:18 (CloseLockDC property)

        palette_id: "PaletteID" field. Palette identifier. Used in toolset only.
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/UTP.cs:57 (PaletteID property)

        Note: UTP shares many fields with UTD (door). See UTD documentation for common fields
        like auto_remove_key, conversation, faction_id, plot, min1_hp, key_required, lockable,
        locked, unlock_dc, key_name, animation_state, maximum_hp, current_hp, hardness,
        fortitude, on_closed, on_damaged, on_death, on_heartbeat, on_lock, on_melee_attack,
        on_open, on_force_power, on_unlock, on_user_defined, static, useable, party_interact,
        on_end_dialog, on_inventory, on_used, comment, description, interruptable, portrait_id,
        trap_detectable, trap_detect_dc, trap_disarmable, trap_disarm_dc, trap_flag,
        trap_one_shot, trap_type, will, on_disarm, on_trap_triggered, bodybag_id, type_id.
    """

    BINARY_TYPE = ResourceType.UTP

    def __init__(  # noqa: PLR0915
        self,
    ):
        self.resref: ResRef = ResRef.from_blank()
        self.conversation: ResRef = ResRef.from_blank()
        self.tag: str = ""
        self.comment: str = ""

        self.name: LocalizedString = LocalizedString.from_invalid()

        self.faction_id: int = 0
        self.appearance_id: int = 0
        self.animation_state: int = 0

        self.auto_remove_key: bool = False
        self.key_name: str = ""
        self.key_required: bool = False
        self.locked: bool = False
        self.lockable: bool = False

        self.unlock_dc: int = 0
        self.unlock_diff: int = 0  # KotOR 2 Only
        self.unlock_diff_mod: int = 0  # KotOR 2 Only

        self.current_hp: int = 0
        self.maximum_hp: int = 0
        self.fortitude: int = 0
        self.hardness: int = 0

        self.min1_hp: bool = False
        self.not_blastable: bool = False
        self.plot: bool = False
        self.static: bool = False
        self.useable: bool = False
        self.party_interact: bool = False

        self.on_closed: ResRef = ResRef.from_blank()
        self.on_damaged: ResRef = ResRef.from_blank()
        self.on_death: ResRef = ResRef.from_blank()
        self.on_end_dialog: ResRef = ResRef.from_blank()
        self.on_open_failed: ResRef = ResRef.from_blank()
        self.on_heartbeat: ResRef = ResRef.from_blank()
        self.on_inventory: ResRef = ResRef.from_blank()
        self.on_melee_attack: ResRef = ResRef.from_blank()
        self.on_force_power: ResRef = ResRef.from_blank()
        self.on_open: ResRef = ResRef.from_blank()
        self.on_lock: ResRef = ResRef.from_blank()
        self.on_unlock: ResRef = ResRef.from_blank()
        self.on_used: ResRef = ResRef.from_blank()
        self.on_user_defined: ResRef = ResRef.from_blank()

        self.has_inventory: bool = False
        self.inventory: list[InventoryItem] = []

        # Deprecated:
        self.description: LocalizedString = LocalizedString.from_invalid()
        self.interruptable: bool = False
        self.portrait_id: int = 0
        self.trap_detectable: bool = False
        self.trap_detect_dc: int = 0
        self.trap_disarmable: bool = False
        self.trap_disarm_dc: int = 0
        self.trap_flag: int = 0
        self.trap_one_shot: bool = False
        self.trap_type: int = 0
        self.will: int = 0
        self.reflex: int = 0
        self.on_disarm: ResRef = ResRef.from_blank()
        self.on_trap_triggered: ResRef = ResRef.from_blank()
        self.bodybag_id: int = 0
        self.type_id: int = 0
        self.palette_id: int = 0
        self.lock_dc: int = 0


def construct_utp(  # noqa: PLR0915
    gff: GFF,
) -> UTP:
    """Constructs a UTP object from a GFF structure.

    Defaults when field missing (from engine): K1 CSWSPlaceable::LoadPlaceable (K1: 0x00585670, TSL: 0x006a1680 legacy PC),
    SavePlaceable (K1: 0x00586a70, TSL: TODO). Tag "", TemplateResRef "", LocName empty; BYTE 0,
    scripts ResRef "". Optional when missing.

    Reference functions: (1) LoadPlaceable root UTP parser, (2) SavePlaceable writer,
    (3) LoadPlaceables area placeables, (4) LoadFromTemplate, (5) CResGFF::ReadField* for Tag,
    LocName, Appearance, HP, ItemList, etc. TSL same semantics; addresses in UTP class References.
    """
    utp = UTP()

    root: GFFStruct = gff.root
    # Identity: Tag "", LocName empty, TemplateResRef "". K1 LoadPlaceable 0x00585670; TSL same (addresses in UTP References). Optional.
    utp.tag = root.acquire("Tag", "")
    utp.name = root.acquire("LocName", LocalizedString.from_invalid())
    utp.resref = root.acquire("TemplateResRef", ResRef.from_blank())
    # Lock/key: AutoRemoveKey, KeyRequired, Lockable, Locked, OpenLockDC, KeyName. Default 0 or "". K1/TSL LoadPlaceable. Optional.
    utp.auto_remove_key = bool(root.acquire("AutoRemoveKey", 0))
    utp.lock_dc = root.acquire("CloseLockDC", 0)
    utp.conversation = root.acquire("Conversation", ResRef.from_blank())
    utp.faction_id = root.acquire("Faction", 0)
    utp.plot = bool(root.acquire("Plot", 0))
    utp.not_blastable = bool(root.acquire("NotBlastable", 0))
    utp.min1_hp = bool(root.acquire("Min1HP", 0))
    utp.key_required = bool(root.acquire("KeyRequired", 0))
    utp.lockable = bool(root.acquire("Lockable", 0))
    utp.locked = bool(root.acquire("Locked", 0))
    utp.unlock_dc = root.acquire("OpenLockDC", 0)
    utp.key_name = root.acquire("KeyName", "")
    utp.animation_state = root.acquire("AnimationState", 0)
    # HP/stats: Appearance, HP, CurrentHP, Hardness, Fort. Default 0. K1/TSL LoadPlaceable. Optional.
    utp.appearance_id = root.acquire("Appearance", 0)
    utp.maximum_hp = root.acquire("HP", 0)
    utp.current_hp = root.acquire("CurrentHP", 0)
    utp.hardness = root.acquire("Hardness", 0)
    utp.fortitude = root.acquire("Fort", 0)
    # Scripts: OnClosed/OnDamaged/OnDeath/OnHeartbeat/OnOpen/OnUsed etc. ResRef "". K1/TSL LoadPlaceable. Optional.
    utp.on_closed = root.acquire("OnClosed", ResRef.from_blank())
    utp.on_damaged = root.acquire("OnDamaged", ResRef.from_blank())
    utp.on_death = root.acquire("OnDeath", ResRef.from_blank())
    utp.on_heartbeat = root.acquire("OnHeartbeat", ResRef.from_blank())
    utp.on_lock = root.acquire("OnLock", ResRef.from_blank())
    utp.on_melee_attack = root.acquire("OnMeleeAttacked", ResRef.from_blank())
    utp.on_open = root.acquire("OnOpen", ResRef.from_blank())
    utp.on_force_power = root.acquire("OnSpellCastAt", ResRef.from_blank())
    utp.on_unlock = root.acquire("OnUnlock", ResRef.from_blank())
    utp.on_user_defined = root.acquire("OnUserDefined", ResRef.from_blank())
    utp.has_inventory = bool(root.acquire("HasInventory", 0))
    utp.party_interact = bool(root.acquire("PartyInteract", 0))
    utp.static = bool(root.acquire("Static", 0))
    utp.useable = bool(root.acquire("Useable", 0))
    utp.on_end_dialog = root.acquire("OnEndDialogue", ResRef.from_blank())
    utp.on_inventory = root.acquire("OnInvDisturbed", ResRef.from_blank())
    utp.on_used = root.acquire("OnUsed", ResRef.from_blank())
    utp.on_open_failed = root.acquire("OnFailToOpen", ResRef.from_blank())
    utp.comment = root.acquire("Comment", "")
    utp.unlock_diff = root.acquire("OpenLockDiff", 0)
    utp.unlock_diff_mod = root.acquire("OpenLockDiffMod", 0)

    # ItemList: InventoryRes "", Dropable 0. K1/TSL LoadPlaceable. Optional.
    item_list: GFFList = root.acquire("ItemList", GFFList())
    for item_struct in item_list:
        resref = item_struct.acquire("InventoryRes", ResRef.from_blank())
        droppable = bool(item_struct.acquire("Dropable", 0))
        utp.inventory.append(InventoryItem(resref, droppable))

    utp.description = root.acquire("Description", LocalizedString.from_invalid())
    utp.interruptable = bool(root.acquire("Interruptable", 0))
    utp.portrait_id = root.acquire("PortraitId", 0)
    utp.trap_detectable = bool(root.acquire("TrapDetectable", 0))
    utp.trap_detect_dc = root.acquire("TrapDetectDC", 0)
    utp.trap_disarmable = bool(root.acquire("TrapDisarmable", 0))
    utp.trap_disarm_dc = root.acquire("DisarmDC", 0)
    utp.trap_flag = root.acquire("TrapFlag", 0)
    utp.trap_one_shot = bool(root.acquire("TrapOneShot", 0))
    utp.trap_type = root.acquire("TrapType", 0)
    utp.will = root.acquire("Will", 0)
    utp.reflex = root.acquire("Ref", 0)
    utp.on_disarm = root.acquire("OnDisarm", ResRef.from_blank())
    utp.on_trap_triggered = root.acquire("OnTrapTriggered", ResRef.from_blank())
    utp.bodybag_id = root.acquire("BodyBag", 0)
    utp.type_id = root.acquire("Type", 0)
    utp.palette_id = root.acquire("PaletteID", 0)

    return utp


def dismantle_utp(  # noqa: PLR0915
    utp: UTP,
    game: Game = Game.K2,
    *,
    use_deprecated: bool = True,
) -> GFF:
    gff = GFF(GFFContent.UTP)

    root: GFFStruct = gff.root
    # Write same defaults as engine read. K1 LoadPlaceable 0x00585670, SavePlaceable 0x00586a70; TSL same (addresses in UTP References). BYTE 0, DWORD 0, CResRef "".
    root.set_string("Tag", utp.tag)
    root.set_locstring("LocName", utp.name)
    root.set_resref("TemplateResRef", utp.resref)
    root.set_uint8("AutoRemoveKey", utp.auto_remove_key)
    root.set_resref("Conversation", utp.conversation)
    root.set_uint32("Faction", utp.faction_id)
    root.set_uint8("Plot", utp.plot)
    root.set_uint8("Min1HP", utp.min1_hp)
    root.set_uint8("KeyRequired", utp.key_required)
    root.set_uint8("Lockable", utp.lockable)
    root.set_uint8("Locked", utp.locked)
    root.set_uint8("OpenLockDC", utp.unlock_dc)
    root.set_string("KeyName", utp.key_name)
    root.set_uint8("AnimationState", utp.animation_state)
    root.set_uint32("Appearance", utp.appearance_id)
    root.set_int16("HP", utp.maximum_hp)
    root.set_int16("CurrentHP", utp.current_hp)
    root.set_uint8("Hardness", utp.hardness)
    root.set_uint8("Fort", utp.fortitude)
    root.set_resref("OnClosed", utp.on_closed)
    root.set_resref("OnDamaged", utp.on_damaged)
    root.set_resref("OnDeath", utp.on_death)
    root.set_resref("OnHeartbeat", utp.on_heartbeat)
    root.set_resref("OnLock", utp.on_lock)
    root.set_resref("OnMeleeAttacked", utp.on_melee_attack)
    root.set_resref("OnOpen", utp.on_open)
    root.set_resref("OnSpellCastAt", utp.on_force_power)
    root.set_resref("OnUnlock", utp.on_unlock)
    root.set_resref("OnUserDefined", utp.on_user_defined)
    root.set_uint8("HasInventory", utp.has_inventory)
    root.set_uint8("PartyInteract", utp.party_interact)
    root.set_uint8("Static", utp.static)
    root.set_uint8("Useable", utp.useable)
    root.set_resref("OnEndDialogue", utp.on_end_dialog)
    root.set_resref("OnInvDisturbed", utp.on_inventory)
    root.set_resref("OnUsed", utp.on_used)
    root.set_string("Comment", utp.comment)

    item_list: GFFList = root.set_list("ItemList", GFFList())
    for i, item in enumerate(utp.inventory):
        item_struct: GFFStruct = item_list.add(i)
        item_struct.set_resref("InventoryRes", item.resref)
        item_struct.set_uint16("Repos_PosX", i)
        item_struct.set_uint16("Repos_PosY", 0)
        if item.droppable:
            item_struct.set_uint8("Dropable", value=True)

    root.set_uint8("PaletteID", utp.palette_id)

    if game.is_k2():
        root.set_uint8("NotBlastable", utp.not_blastable)
        root.set_uint8("OpenLockDiff", utp.unlock_diff)
        root.set_int8("OpenLockDiffMod", utp.unlock_diff_mod)
        root.set_resref("OnFailToOpen", utp.on_open_failed)
        root.set_uint8("CloseLockDC", utp.lock_dc)

    if use_deprecated:
        root.set_locstring("Description", utp.description)
        root.set_uint8("Interruptable", utp.interruptable)
        root.set_uint16("PortraitId", utp.portrait_id)
        root.set_uint8("TrapDetectable", utp.trap_detectable)
        root.set_uint8("TrapDetectDC", utp.trap_detect_dc)
        root.set_uint8("TrapDisarmable", utp.trap_disarmable)
        root.set_uint8("DisarmDC", utp.trap_disarm_dc)
        root.set_uint8("TrapFlag", utp.trap_flag)
        root.set_uint8("TrapOneShot", utp.trap_one_shot)
        root.set_uint8("TrapType", utp.trap_type)
        root.set_uint8("Will", utp.will)
        root.set_uint8("Ref", utp.reflex)
        root.set_resref("OnDisarm", utp.on_disarm)
        root.set_resref("OnTrapTriggered", utp.on_trap_triggered)
        root.set_uint8("BodyBag", utp.bodybag_id)
        root.set_uint8("Type", utp.type_id)

    return gff


def read_utp(
    source: SOURCE_TYPES,
    offset: int = 0,
    size: int | None = None,
) -> UTP:
    gff: GFF = read_gff(source, offset, size)
    return construct_utp(gff)


def write_utp(
    utp: UTP,
    target: TARGET_TYPES,
    game: Game = Game.K2,
    file_format: ResourceType = ResourceType.GFF,
    *,
    use_deprecated: bool = True,
):
    gff: GFF = dismantle_utp(utp, game, use_deprecated=use_deprecated)
    write_gff(gff, target, file_format)


def bytes_utp(
    utp: UTP,
    game: Game = Game.K2,
    file_format: ResourceType = ResourceType.GFF,
    *,
    use_deprecated: bool = True,
) -> bytes:
    gff: GFF = dismantle_utp(utp, game, use_deprecated=use_deprecated)
    return bytes_gff(gff, file_format)



from __future__ import annotations

from typing import TYPE_CHECKING

from qtpy.QtCore import QTimer, Qt
from qtpy.QtWidgets import (
    QAbstractItemView,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QLabel,
    QListWidget,
    QListWidgetItem,
    QVBoxLayout,
)

if TYPE_CHECKING:
    from qtpy.QtWidgets import QWidget


class SetDefaultColumnsDialog(QDialog):
    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)
        self.setWindowTitle("Set Default Columns")
        self.setMinimumSize(400, 500)

        layout = QVBoxLayout(self)

        # Folder type selection
        folder_type_layout = QHBoxLayout()
        folder_type_layout.addWidget(QLabel("Folder type:"))
        self.folder_type_combo: QComboBox = QComboBox()
        self.folder_type_combo.addItem("General")
        folder_type_layout.addWidget(self.folder_type_combo)
        layout.addLayout(folder_type_layout)

        # Columns list
        self.columns_list: QListWidget = QListWidget()
        self.columns_list.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)
        self.columns_list.setAcceptDrops(True)
        self.columns_list.setDragEnabled(True)
        self.columns_list.setDropIndicatorShown(True)
        self.columns_list.setSelectionMode(QAbstractItemView.SelectionMode.NoSelection)
        self.populate_columns()
        layout.addWidget(self.columns_list)

        # Description
        self.description_label: QLabel = QLabel("Description: File name")
        layout.addWidget(self.description_label)

        # OK and Cancel buttons
        button_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        layout.addWidget(button_box)

        # Connect signals
        self.columns_list.itemClicked.connect(self.update_description)

    def populate_columns(self):
        columns = [
            "Name",
            "Type",
            "Size",
            "Date Modified",
            "Attributes",
            "Size On Disk",
            "8.3 Name",
            "Owner",
            "Product Name",
            "Company",
            "Description",
            "File Version",
            "Product Version",
            "Shortcut to",
            "Hard links",
            "Extension",
            "Date Created",
            "Date Accessed",
            "Title",
            "Subject",
            "Authors",
            "Keywords",
            "Comment",
            "Camera Model",
            "Date Taken",
            "Width",
            "Height",
            "Bit rate",
            "Copyright",
            "Duration",
            "Protected",
            "Rating",
            "Album artist",
            "Album",
            "Beats-per-minute",
            "Composer",
            "Conductor",
            "Director",
            "Genre",
            "Language",
            "Broadcast date",
            "Channel",
            "Station name",
            "Mood",
            "Parental rating",
            "Parental rating reason",
            "Period",
            "Producer",
        ]
        for column in columns:
            item = QListWidgetItem(column, self.columns_list, QListWidgetItem.ItemType.UserType)
            item.setCheckState(Qt.CheckState.Unchecked)
            item.setFlags(item.flags() | Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsDragEnabled)
            self.columns_list.addItem(item)

    def update_description(self, item: QListWidgetItem):
        self.description_label.setText(f"Description: {item.text()}")

    def get_selected_columns(self) -> list[str]:
        selected_columns: list[str] = []
        for i in range(self.columns_list.count()):
            item: QListWidgetItem | None = self.columns_list.item(i)
            if item is not None and item.checkState() == Qt.CheckState.Checked:
                selected_columns.append(item.text())
        return selected_columns


if __name__ == "__main__":
    import sys

    from qtpy.QtWidgets import QApplication

    app = QApplication(sys.argv)
    dialog = SetDefaultColumnsDialog()

    def on_accepted(*args, **kwargs):
        selected_columns = dialog.get_selected_columns()
        print("Selected columns:", selected_columns)
        QTimer.singleShot(0, app.quit)

    dialog.accepted.connect(on_accepted)
    dialog.show()
    sys.exit(app.exec())



"""Door (UTD) model and resource resolution using genericdoors.2da and installation."""

from __future__ import annotations

from typing import TYPE_CHECKING

from loggerplus import RobustLogger
from pykotor.extract.file import ResourceIdentifier
from pykotor.extract.installation import SearchLocation
from pykotor.resource.formats.mdl import read_mdl
from pykotor.resource.formats.tpc import read_tpc
from pykotor.resource.formats.twoda import TwoDA, read_2da
from pykotor.resource.generics.utd import read_utd
from pykotor.resource.type import ResourceType
from pykotor.tools.model import iterate_textures
from pykotor.tools.resource_lookup import load_2da_with_fallback, read_location_data
from utility.common.geometry import Vector3

if TYPE_CHECKING:
    from pykotor.common.module import Module
    from pykotor.extract.installation import Installation
    from pykotor.resource.formats.mdl import MDL, MDLNode
    from pykotor.resource.generics.utd import UTD
    from pykotor.resource.type import SOURCE_TYPES


def get_model(
    utd: UTD,
    installation: Installation,
    *,
    genericdoors: TwoDA | SOURCE_TYPES | None = None,
) -> str:
    """Returns the model name for the given door.

    References:
    ----------

    Derivations and Other Implementations:
    -------------------------------------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/module/ModuleDoor.ts (Door appearance handling)


        Note: Door model lookup uses genericdoors.2da


    If no value is specified for the genericdoor parameters then it will be loaded from the given installation.

    Args:
    ----
        utd: UTD object of the door to lookup the model for.
        installation: The relevant installation.
        genericdoors: The genericdoors.2da loaded into a TwoDA object.

    Returns:
    -------
        Returns the model name for the door.

    Raises:
    ------
        ValueError: genericdoors.2da not found in passed arguments OR the installation.
    """
    if genericdoors is None:
        loaded = load_2da_with_fallback(installation, "genericdoors", RobustLogger())
        if loaded is None:
            raise ValueError("Resource 'genericdoors.2da' not found in the installation, cannot get UTD model.")
        genericdoors = loaded
    if not isinstance(genericdoors, TwoDA):
        genericdoors = read_2da(genericdoors)

    return genericdoors.get_row(utd.appearance_id).get_string("modelname")


def load_genericdoors_2da(
    installation: Installation,
    logger: RobustLogger | None = None,
) -> TwoDA | None:
    """Load genericdoors.2da from installation using priority order.

    Tries locations() first (more reliable), then falls back to resource().
    Searches in Override first, then Chitin.

    Args:
    ----
        installation: The game installation instance
        logger: Optional logger for debugging

    Returns:
    -------
        TwoDA object if found, None otherwise
    """
    if logger is None:
        logger = RobustLogger()
    return load_2da_with_fallback(installation, "genericdoors", logger)


def extract_door_walkmeshes(
    utd_data: bytes,
    installation: Installation,
    *,
    module: Module | None = None,
    logger: RobustLogger | None = None,
) -> dict[str, bytes]:
    """Extract door walkmeshes (DWK files) for a door.

    Doors have 3 walkmesh states: closed (0), open1 (1), open2 (2)
    Format: <modelname>0.dwk, <modelname>1.dwk, <modelname>2.dwk

    References:
    ----------

    Derivations and Other Implementations:
    -------------------------------------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/module/ModuleDoor.ts:992



    Args:
    ----
        utd_data: UTD door data bytes
        installation: The game installation instance
        module: Optional Module instance to search for DWK files in module resources first
        logger: Optional logger for debugging

    Returns:
    -------
        Dictionary mapping dwk_key ("dwk0", "dwk1", "dwk2") to DWK file data bytes
    """
    if logger is None:
        logger = RobustLogger()

    door_walkmeshes: dict[str, bytes] = {}

    try:
        utd = read_utd(utd_data)

        # Get door model name from UTD using genericdoors.2da
        genericdoors_2da = load_genericdoors_2da(installation, logger)
        if not genericdoors_2da:
            logger.debug("Could not load genericdoors.2da, cannot extract door walkmeshes")
            return door_walkmeshes

        door_model_name = get_model(utd, installation, genericdoors=genericdoors_2da)
        if not door_model_name:
            logger.debug(f"Could not get model name for door (appearance_id={utd.appearance_id})")
            return door_walkmeshes

        # Try to extract DWK files: modelname0.dwk, modelname1.dwk, modelname2.dwk
        dwk_variants = [
            (f"{door_model_name}0", "dwk0"),
            (f"{door_model_name}1", "dwk1"),
            (f"{door_model_name}2", "dwk2"),
        ]

        for dwk_resname, dwk_key in dwk_variants:
            try:
                # Try to find DWK in module resources first (if module provided)
                if module is not None:
                    dwk_resource = module.resource(dwk_resname, ResourceType.DWK)
                    if dwk_resource is not None:
                        dwk_data = dwk_resource.data()
                        if dwk_data:
                            door_walkmeshes[dwk_key] = dwk_data
                            logger.debug(f"Found DWK '{dwk_resname}' (state: {dwk_key}) from module")
                            continue

                # Try installation locations
                dwk_locations = installation.locations(
                    [ResourceIdentifier(resname=dwk_resname, restype=ResourceType.DWK)],
                    [
                        SearchLocation.OVERRIDE,
                        SearchLocation.MODULES,
                        SearchLocation.CHITIN,
                    ],
                )
                for dwk_ident, dwk_loc_list in dwk_locations.items():
                    if dwk_loc_list:
                        dwk_loc = dwk_loc_list[0]
                        dwk_data = read_location_data(dwk_loc)
                        if dwk_data is None:
                            continue
                        door_walkmeshes[dwk_key] = dwk_data
                        logger.debug(f"Found DWK '{dwk_resname}' (state: {dwk_key}) from installation")
                        break
            except Exception:  # noqa: BLE001
                # DWK variant not found, skip it
                pass

    except Exception as e:  # noqa: BLE001
        logger.debug(f"Could not extract DWK walkmeshes: {e}")

    return door_walkmeshes


def _get_model_variations(model_name: str) -> list[str]:
    """Get list of model name variations to try when searching for resources.

    Some doors have models that don't exist, so we try various name formats.

    Args:
    ----
        model_name: Original model name

    Returns:
    -------
        List of model name variations (original case, lowercase, uppercase, normalized)
    """
    variations = [
        model_name,  # Original case
        model_name.lower(),  # Lowercase
        model_name.upper(),  # Uppercase
        model_name.lower().replace(".mdl", "").replace(".mdx", ""),  # Normalized lowercase
    ]

    # Remove duplicates while preserving order
    seen = set()
    return [v for v in variations if v not in seen and not seen.add(v)]


def _load_mdl_with_variations(
    model_name: str,
    installation: Installation,
    logger: RobustLogger | None = None,
) -> tuple[MDL | None, bytes | None]:
    """Load MDL file trying multiple name variations.

    Args:
    ----
        model_name: Base model name to try
        installation: The game installation instance
        logger: Optional logger for debugging

    Returns:
    -------
        Tuple of (MDL object, MDL data bytes) if found, (None, None) otherwise
    """
    if logger is None:
        logger = RobustLogger()

    model_variations = _get_model_variations(model_name)

    # Try locations() first (more reliable, searches multiple locations)
    for model_var in model_variations:
        try:
            location_results = installation.locations(
                [ResourceIdentifier(resname=model_var, restype=ResourceType.MDL)],
                [
                    SearchLocation.OVERRIDE,
                    SearchLocation.MODULES,
                    SearchLocation.CHITIN,
                ],
            )
            for res_ident, loc_list in location_results.items():
                if loc_list:
                    loc = loc_list[0]
                    try:
                        mdl_data = read_location_data(loc)
                        if mdl_data is None:
                            continue
                        mdl = read_mdl(mdl_data)
                        return mdl, mdl_data
                    except Exception:  # noqa: BLE001
                        continue
        except Exception:  # noqa: BLE001
            continue

    # Fallback to resource() if locations() didn't work
    for model_var in model_variations:
        try:
            mdl_result = installation.resource(model_var, ResourceType.MDL)
            if mdl_result and mdl_result.data:
                mdl = read_mdl(mdl_result.data)
                return mdl, mdl_result.data
        except Exception:  # noqa: BLE001
            continue

    return None, None


def _get_door_dimensions_from_model(
    mdl: MDL,
    model_name: str,
    door_name: str | None = None,
    logger: RobustLogger | None = None,
) -> tuple[float, float] | None:
    """Calculate door dimensions from MDL bounding box.

    Doors are typically oriented along Y axis (width) and Z axis (height).
    X is typically depth/thickness.

    Args:
    ----
        mdl: MDL model object
        model_name: Model name for logging
        door_name: Optional door name for logging
        logger: Optional logger for debugging

    Returns:
    -------
        Tuple of (width, height) if calculated successfully, None otherwise
    """
    if logger is None:
        logger = RobustLogger()

    if not mdl or not mdl.root:
        return None

    bb_min = Vector3(1000000, 1000000, 1000000)
    bb_max = Vector3(-1000000, -1000000, -1000000)

    # Iterate through all nodes and their meshes
    nodes_to_check: list[MDLNode] = [mdl.root]
    mesh_count = 0
    while nodes_to_check:
        node: MDLNode = nodes_to_check.pop()
        if node.mesh:
            mesh_count += 1
            # Use mesh bounding box if available
            if node.mesh.bb_min and node.mesh.bb_max:
                bb_min.x = min(bb_min.x, node.mesh.bb_min.x)
                bb_min.y = min(bb_min.y, node.mesh.bb_min.y)
                bb_min.z = min(bb_min.z, node.mesh.bb_min.z)
                bb_max.x = max(bb_max.x, node.mesh.bb_max.x)
                bb_max.y = max(bb_max.y, node.mesh.bb_max.y)
                bb_max.z = max(bb_max.z, node.mesh.bb_max.z)
            # Fallback: calculate from vertex positions if bounding box not set
            elif node.mesh.vertex_positions:
                for vertex in node.mesh.vertex_positions:
                    bb_min.x = min(bb_min.x, vertex.x)
                    bb_min.y = min(bb_min.y, vertex.y)
                    bb_min.z = min(bb_min.z, vertex.z)
                    bb_max.x = max(bb_max.x, vertex.x)
                    bb_max.y = max(bb_max.y, vertex.y)
                    bb_max.z = max(bb_max.z, vertex.z)

        # Check child nodes
        nodes_to_check.extend(node.children)

    # Calculate dimensions from bounding box
    # Width is typically the Y dimension (horizontal when door is closed)
    # Height is typically the Z dimension (vertical)
    if bb_min.x < 1000000:  # Valid bounding box calculated
        width = abs(bb_max.y - bb_min.y)
        height = abs(bb_max.z - bb_min.z)

        # Only use calculated values if they're reasonable (not zero or extremely large)
        if 0.1 < width < 50.0 and 0.1 < height < 50.0:
            door_name_str = f"'{door_name}'" if door_name else ""
            logger.debug(f"[DOOR DEBUG] Extracted dimensions for door {door_name_str}: {width:.2f} x {height:.2f} (from {mesh_count} meshes, model='{model_name}')")
            return width, height
        else:
            door_name_str = f"'{door_name}'" if door_name else ""
            logger.warning(f"Calculated dimensions for door {door_name_str} out of range: {width:.2f} x {height:.2f}, using defaults")
    else:
        door_name_str = f"'{door_name}'" if door_name else ""
        logger.warning(f"Could not calculate bounding box for door {door_name_str} (processed {mesh_count} meshes), using defaults")

    return None


def _get_door_dimensions_from_texture(
    model_name: str,
    installation: Installation,
    door_name: str | None = None,
    logger: RobustLogger | None = None,
) -> tuple[float, float] | None:
    """Calculate door dimensions from door texture as fallback.

    Typical door textures are 256x512 or 512x1024 pixels.
    Typical door dimensions are 2-6 units wide, 2.5-3.5 units tall.
    Assuming 1 pixel ≈ 0.008-0.01 world units for doors.

    Args:
    ----
        model_name: Model name to get textures from
        installation: The game installation instance
        door_name: Optional door name for logging
        logger: Optional logger for debugging

    Returns:
    -------
        Tuple of (width, height) if calculated successfully, None otherwise
    """
    if logger is None:
        logger = RobustLogger()

    # Get textures from the model
    texture_names: list[str] = []
    model_variations = _get_model_variations(model_name)

    for model_var in model_variations:
        try:
            mdl_result = installation.resource(model_var, ResourceType.MDL)
            if mdl_result and mdl_result.data:
                texture_names = list(iterate_textures(mdl_result.data))
                break
        except Exception:  # noqa: BLE001
            continue

    if not texture_names:
        return None

    # Try to load the first texture
    texture_name = texture_names[0]
    texture_result = installation.resource(texture_name, ResourceType.TPC)
    if not texture_result:
        # Try TGA as fallback
        texture_result = installation.resource(texture_name, ResourceType.TGA)

    if not texture_result or not texture_result.data:
        return None

    # Read texture to get dimensions
    tex_width = 0
    tex_height = 0

    if texture_result.restype == ResourceType.TPC:
        tpc = read_tpc(texture_result.data)
        tex_width, tex_height = tpc.dimensions()
    elif texture_result.restype == ResourceType.TGA:
        # TGA header: width at offset 12, height at offset 14 (little-endian)
        if len(texture_result.data) >= 18:
            tex_width = int.from_bytes(texture_result.data[12:14], "little")
            tex_height = int.from_bytes(texture_result.data[14:16], "little")

    if tex_width <= 0 or tex_height <= 0:
        return None

    # Convert texture pixels to world units
    # Use aspect ratio to determine which dimension is width vs height
    # Doors are typically taller than wide, so height > width
    if tex_height > tex_width:
        # Portrait orientation - height is vertical, width is horizontal
        # Typical: 256x512 = 2.0x4.0, 512x1024 = 4.0x8.0
        # Scale factor: ~0.008-0.01 units per pixel
        scale_factor = 0.008  # Conservative estimate
        door_width = tex_width * scale_factor
        door_height = tex_height * scale_factor
    else:
        # Landscape or square - assume standard door proportions
        # Use height as the primary dimension
        scale_factor = 0.008
        door_height = tex_height * scale_factor
        # Width is typically 0.6-0.8x height for doors
        door_width = door_height * 0.7

    # Clamp to reasonable values
    door_width = max(1.0, min(door_width, 10.0))
    door_height = max(1.5, min(door_height, 10.0))

    return door_width, door_height


def get_door_dimensions(
    utd_data: bytes,
    installation: Installation,
    *,
    door_name: str | None = None,
    default_width: float = 2.0,
    default_height: float = 3.0,
    genericdoors: TwoDA | None = None,
    logger: RobustLogger | None = None,
) -> tuple[float, float]:
    """Get door dimensions (width, height) from model or texture.

    Tries to extract dimensions from MDL bounding box first, then falls back
    to texture-based estimation if model extraction fails.

    References:
    ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        Door dimension calculation logic from kit.py extract_kit()


    Args:
    ----
        utd_data: UTD door data bytes
        installation: The game installation instance
        door_name: Optional door name for logging
        default_width: Default width if extraction fails (default: 2.0)
        default_height: Default height if extraction fails (default: 3.0)
        logger: Optional logger for debugging

    Returns:
    -------
        Tuple of (width, height) in world units
    """
    if logger is None:
        logger = RobustLogger()

    door_width = default_width
    door_height = default_height
    door_name_str = f"'{door_name}'" if door_name else ""

    try:
        utd = read_utd(utd_data)
        logger.debug(f"[DOOR DEBUG] Processing door {door_name_str} (appearance_id={utd.appearance_id})")

        # Get door model name from UTD using genericdoors.2da
        # Use pre-loaded genericdoors_2da if provided, otherwise load it
        genericdoors_2da = genericdoors if genericdoors is not None else load_genericdoors_2da(installation, logger)
        if not genericdoors_2da:
            logger.warning(f"Could not load genericdoors.2da for door {door_name_str}, using defaults")
            return door_width, door_height

        model_name = get_model(utd, installation, genericdoors=genericdoors_2da)
        if not model_name:
            logger.warning(f"Could not get model name for door {door_name_str} (appearance_id={utd.appearance_id}), using defaults")
            return door_width, door_height

        # Try method 1: Get dimensions from model bounding box
        mdl, mdl_data = _load_mdl_with_variations(model_name, installation, logger)
        if mdl:
            dimensions = _get_door_dimensions_from_model(mdl, model_name, door_name, logger)
            if dimensions:
                door_width, door_height = dimensions
                return door_width, door_height
            else:
                logger.warning(f"Could not extract dimensions from model '{model_name}' for door {door_name_str}, trying texture fallback")
        else:
            model_variations = _get_model_variations(model_name)
            logger.warning(
                f"Could not load MDL '{model_name}' (tried variations: {model_variations}) "
                f"for door {door_name_str} "
                f"(appearance_id={utd.appearance_id}), trying texture fallback"
            )

        # Fallback: Get dimensions from door texture if model-based extraction failed
        dimensions = _get_door_dimensions_from_texture(model_name, installation, door_name, logger)
        if dimensions:
            door_width, door_height = dimensions
        else:
            logger.debug(f"[DOOR DEBUG] Door {door_name_str}: Using default dimensions ({default_width} x {default_height}) - model and texture extraction failed")

    except Exception as e:  # noqa: BLE001
        logger.warning(f"Failed to get dimensions for door {door_name_str}: {e}")

    logger.debug(f"[DOOR DEBUG] Final dimensions for door {door_name_str}: width={door_width:.2f}, height={door_height:.2f}")
    return door_width, door_height



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from typing_extensions import Self


def is_string_like(obj: Any) -> bool:  # sourcery skip: use-fstring-for-concatenation
    try:
        _ = obj + ""
    except Exception:  # pylint: disable=W0718  # noqa: BLE001
        return False
    else:
        return True


class WrappedStr(str):
    __slots__: tuple[str, ...] = ("_content",)

    def __init__(
        self,
        content: Self | str = "",
    ):
        if content is None:
            msg = f"Cannot initialize {self.__class__.__name__}(None), expected a str-like argument"
            raise RuntimeError(msg)
        if isinstance(content, WrappedStr):
            content = content._content  # noqa: SLF001
        self._content: str = content

    def __repr__(self):
        return f"WrappedStr({self._content!r})"

    def __reduce__(self):
        return (self.__class__, (self._content,))

    if not TYPE_CHECKING:  # Prevents pylance in vs code from bringing us here in 'go to definition'

        def __getattribute__(self, name: str):
            try:
                return super().__getattribute__(name)
            except AttributeError:
                # Forward attribute access to _content - legitimate use of getattr for delegation
                try:
                    return getattr(self._content, name)
                except AttributeError:
                    raise AttributeError(f"'{type(self).__name__}' object and its '_content' attribute have no attribute '{name}'")

    # region Forwards Compatibility
    # Check if str has __reduce_ex__ - legitimate runtime check for optional builtin method
    if not hasattr(str, "__reduce_ex__"):

        def __reduce_ex__(self, protocol: int):
            if protocol >= 2:  # Protocol version 2 or higher uses a more efficient pickling format  # noqa: PLR2004
                return (self.__class__, (str(self),), None, None, None)
            return self.__reduce__()
    # endregion

    # region Backwards Compatibility
    @classmethod
    def _assert_str_or_none_type(
        cls: type[Self],
        var: Any,
    ) -> str:
        if var is None:
            return None  # type: ignore[return-value]
        if not isinstance(var, (cls, str)):
            msg = f"Expected str-like, got '{var}' of type {var.__class__}"
            raise TypeError(msg)
        return str(var)

    # Check if str has removeprefix - legitimate runtime check for optional builtin method
    if not hasattr(str, "removeprefix"):

        def removeprefix(
            self,
            __prefix: WrappedStr | str,
        ) -> Self:
            parsed_prefix: str = self._assert_str_or_none_type(__prefix)
            if self._content.startswith(parsed_prefix):
                return self.__class__(self._content[len(parsed_prefix) :])
            return self.__class__(self._content)

    # Check if str has removesuffix - legitimate runtime check for optional builtin method
    if not hasattr(str, "removesuffix"):

        def removesuffix(
            self,
            __suffix: WrappedStr | str,
        ) -> Self:
            parsed_suffix: str = self._assert_str_or_none_type(__suffix)
            if self._content.endswith(parsed_suffix):
                return self.__class__(self._content[: -len(parsed_suffix)])
            return self.__class__(self._content)

    def __getstate__(self) -> str:
        return self._content

    # endregion



"""Central restype -> decoder registry for resource bytes -> domain objects.

Reused by Installation, ModuleResource, and InstallationWorkspace so decode
logic lives in one place. Decoders accept bytes and return the typed domain object
(e.g. read_utc(data) -> UTC).
"""

from __future__ import annotations

from typing import Any, Callable

from pykotor.resource.formats.bwm import read_bwm
from pykotor.resource.formats.gff import read_gff
from pykotor.resource.formats.lyt import read_lyt
from pykotor.resource.formats.tpc import read_tpc
from pykotor.resource.formats.vis import read_vis
from pykotor.resource.generics.are import read_are
from pykotor.resource.generics.dlg import read_dlg
from pykotor.resource.generics.git import read_git
from pykotor.resource.generics.ifo import read_ifo
from pykotor.resource.generics.pth import read_pth
from pykotor.resource.generics.utc import read_utc
from pykotor.resource.generics.utd import read_utd
from pykotor.resource.generics.ute import read_ute
from pykotor.resource.generics.uti import read_uti
from pykotor.resource.generics.utm import read_utm
from pykotor.resource.generics.utp import read_utp
from pykotor.resource.generics.uts import read_uts
from pykotor.resource.generics.utt import read_utt
from pykotor.resource.generics.utw import read_utw
from pykotor.resource.type import ResourceType

# Decoder: bytes -> domain object (GFF, UTC, LYT, etc.)
_DECODER_REGISTRY: dict[ResourceType, Callable[[bytes], Any]] = {
    ResourceType.ARE: read_are,
    ResourceType.DLG: read_dlg,
    ResourceType.GIT: read_git,
    ResourceType.IFO: read_ifo,
    ResourceType.LYT: read_lyt,
    ResourceType.NCS: lambda data: data,
    ResourceType.PTH: read_pth,
    ResourceType.TPC: read_tpc,
    ResourceType.TGA: read_tpc,
    ResourceType.UTD: read_utd,
    ResourceType.UTE: read_ute,
    ResourceType.UTI: read_uti,
    ResourceType.UTM: read_utm,
    ResourceType.UTP: read_utp,
    ResourceType.UTS: read_uts,
    ResourceType.UTT: read_utt,
    ResourceType.UTW: read_utw,
    ResourceType.UTC: read_utc,
    ResourceType.VIS: read_vis,
    ResourceType.WOK: read_bwm,
    ResourceType.GFF: read_gff,
}


def get_decoder(restype: ResourceType) -> Callable[[bytes], Any] | None:
    """Return the decoder for the given ResourceType, or None if none is registered."""
    return _DECODER_REGISTRY.get(restype)



"""Backward compatibility: utility.ui_libraries mirrors utility.gui so legacy imports resolve.

Legacy: from utility.ui_libraries.qt.widgets.theme.theme_dialog import ThemeDialog
Resolves to: utility.gui.qt.widgets.theme.theme_dialog (same namespace).
"""
from utility import gui
import sys

# Replace this package with gui so utility.ui_libraries.qt.widgets... resolves to gui.qt.widgets...
sys.modules["utility.ui_libraries"] = gui



from __future__ import annotations

import ast
import builtins
import json
import os
import pathlib
import re
import sys
import typing

from abc import ABC
from contextlib import suppress
from dataclasses import MISSING, dataclass, fields
from pathlib import Path, PurePath
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    TypeVar,
    Union,
    get_args,
    get_origin,
    overload,
)
from urllib.parse import urlparse

from utility.misc import ensure_directory_exists

if TYPE_CHECKING:
    from typing_extensions import Self


if __name__ == "__main__":
    with suppress(Exception):

        def update_sys_path(path: pathlib.Path):
            working_dir = str(path)
            if working_dir not in sys.path:
                sys.path.append(working_dir)

        file_absolute_path = pathlib.Path(__file__).resolve()

        pykotor_path = file_absolute_path.parents[6] / "Libraries" / "PyKotor" / "src" / "pykotor"
        if pykotor_path.exists():
            update_sys_path(pykotor_path.parent)
        pykotor_gl_path = file_absolute_path.parents[6] / "Libraries" / "PyKotorGL" / "src" / "pykotor"
        if pykotor_gl_path.exists():
            update_sys_path(pykotor_gl_path.parent)
        utility_path = file_absolute_path.parents[6] / "Libraries" / "Utility" / "src"
        if utility_path.exists():
            update_sys_path(utility_path)
        toolset_path = file_absolute_path.parents[5] / "Tools/HolocronToolset/src/toolset"
        if toolset_path.exists():
            update_sys_path(toolset_path.parent)
            if __name__ == "__main__":
                os.chdir(toolset_path)

T = TypeVar("T")


class AbstractAPIResult(ABC):  # noqa: B024
    _type_cache: ClassVar[dict] = {}

    def __getitem__(self, attr_name: str):
        return self.__dict__[attr_name]

    def __setitem__(self, attr_name: str, value: Any):
        self.__dict__[attr_name] = value

    @classmethod
    def from_dict(cls, json_dict: dict[str, Any]) -> Self:
        assert isinstance(json_dict, dict), f"type {json_dict.__class__.__name__} with contents {json_dict}"
        processed_data = {}
        for this_field in fields(cls):
            key = this_field.name
            expected_type = this_field.metadata.get("default_type", (this_field.type,))
            if key not in json_dict:
                # print(f"{cls.__name__} Warning: Missing expected key '{key}' in data")
                processed_data[key] = this_field.default if this_field.default is not MISSING else None
            else:
                value = json_dict[key]
                processed_data[key] = cls.handle_casting(value, expected_type)

        # for key, value in json_dict.items():
        # if key not in processed_data:
        #    print(f"{cls.__name__} Warning: Unexpected key '{key}' in data with value '{value!r}'")

        return cls(**processed_data)

    def to_dict(self) -> dict[str, Any]:
        result = {}
        for field in fields(self):
            value = getattr(self, field.name)
            if isinstance(value, AbstractAPIResult):
                result[field.name] = value.to_dict()
            elif isinstance(value, list):
                result[field.name] = [item.to_dict() if isinstance(item, AbstractAPIResult) else item for item in value]
            elif isinstance(value, dict):
                result[field.name] = {k: v.to_dict() if isinstance(v, AbstractAPIResult) else v for k, v in value.items()}
            else:
                result[field.name] = value
        return result

    @classmethod
    def resolve_type(cls, type_str: str):
        # sourcery skip: assign-if-exp, reintroduce-else
        def resolve(node):
            if isinstance(node, ast.Name):
                # Map 'list', 'dict' to typing versions when used in type annotations
                if node.id == "list":
                    return list
                if node.id == "dict":
                    return dict
                if node.id == "None":
                    return type(None)
                # Check builtins and typing module for other types
                if hasattr(builtins, node.id):
                    return getattr(builtins, node.id)
                if hasattr(typing, node.id):
                    return getattr(typing, node.id)
                # Check current module's namespace
                if node.id in globals():
                    return globals()[node.id]
                # Check imported modules in current context
                if node.id in sys.modules:
                    return sys.modules[node.id]
            elif isinstance(node, ast.Subscript):
                base = resolve(node.value)
                if isinstance(node.slice, ast.Index):  # Python 3.8 and earlier
                    index = resolve(node.slice.value)
                elif isinstance(node.slice, ast.Tuple):  # Multiple type parameters
                    index = tuple(resolve(i) for i in node.slice.elts)
                else:
                    index = resolve(node.slice)
                # Handle 'None' within subscript parameters
                if isinstance(index, tuple):
                    index = tuple(i if i is not None else type(None) for i in index)
                    return base[index]
                # Handle the case where the entire index is None
                if index is None:
                    return typing.Union[base, type(None)]
                return base[index]
            elif isinstance(node, ast.BinOp) and isinstance(node.op, ast.BitOr):
                # Handle union types like 'int | None'
                left = resolve(node.left)
                right = resolve(node.right)
                return typing.Union[left, right]
            elif isinstance(node, ast.Constant) and node.value is None:
                return type(None)
            return None

        tree = ast.parse(type_str, mode="eval")
        assert not isinstance(tree, str), "Parsed tree should not be a string."
        resolved_type = resolve(tree.body)
        cls._type_cache[type_str] = resolved_type
        return resolved_type

    @classmethod
    def handle_casting(cls, value: Any, expected_types: tuple[type[T], ...]) -> T:
        # Optimize by caching converted types
        converted_types = tuple(cls._type_cache.get(t, cls.resolve_type(t) if isinstance(t, str) else t) for t in expected_types)
        cls._type_cache.update(zip(expected_types, converted_types))

        # Convert all string type names to actual types
        converted_types = []
        for this_type in expected_types:
            if isinstance(this_type, str):
                converted_types.append(cls.resolve_type(this_type))
            else:
                converted_types.append(this_type)
        converted_types = tuple(converted_types)

        if value is None:
            for this_type in converted_types:
                origin = get_origin(this_type) if hasattr(this_type, "__origin__") else this_type
                args = get_args(this_type)
                if origin is None:
                    return None
                if type(None) in args:
                    return None
            raise TypeError(f"Expected value of type(s) {converted_types}, got value of None")

        for this_type in converted_types:
            if this_type is typing.Any:
                return value
            origin = get_origin(this_type) if hasattr(this_type, "__origin__") else this_type
            args = get_args(this_type)

            if origin is Union:
                # Handle Union types
                for arg in args:
                    try:
                        return cls.handle_casting(value, (arg,))
                    except (TypeError, ValueError):  # noqa: S112, PERF203
                        continue
                raise ValueError(f"Value {value} does not match any type in {expected_types}")

            if origin and issubclass(origin, str):
                if isinstance(value, str):
                    return value
                raise ValueError(f"Expected value of type {origin}, got {type(value)}")

            if issubclass(origin, typing.Mapping):
                if not value and isinstance(value, origin):  # Empty check
                    return value
                # Handle mapping types
                key_type, value_type = args if len(args) == 2 else (Any, Any)
                return {cls.handle_casting(k, (key_type,)): cls.handle_casting(v, (value_type,)) for k, v in value.items()}

            if issubclass(origin, typing.Iterable):
                if not value and isinstance(value, origin):  # Empty check
                    return value
                # Handle iterable types
                item_type = args[0] if args else Any
                return [cls.handle_casting(item, (item_type,)) for item in value]

            # Handle non-generic types
            if isinstance(value, origin):
                return value
            if issubclass(origin, AbstractAPIResult):
                return origin.from_dict(value)

            # Attempt to convert the value to the expected type
            return origin(value)

        raise ValueError(f"Expected value of type {expected_types}, got {type(value)} with data '{value}'")


@dataclass
class Asset(AbstractAPIResult):
    url: str = ""
    id: int = -1
    name: str = ""
    label: str | None = None
    state: str = ""
    content_type: str = ""
    size: int = -1
    download_count: int = -1
    created_at: str = ""
    updated_at: str = ""
    browser_download_url: str = ""
    node_id: int = -1
    uploader: str = ""


@dataclass
class LinksData(AbstractAPIResult):
    self: str
    git: str | None
    html: str


@dataclass
class ContentInfoData(AbstractAPIResult):
    name: str
    path: str
    sha: str
    size: int
    url: str
    html_url: str
    git_url: str
    download_url: str | None
    type: str
    _links: LinksData


@dataclass
class GithubRelease:
    url: str
    assets_url: str
    upload_url: str
    html_url: str
    id: int
    author: dict
    node_id: str
    tag_name: str
    target_commitish: str
    name: str
    draft: bool
    prerelease: bool
    created_at: str
    published_at: str
    assets: list[Asset]
    tarball_url: str
    zipball_url: str
    body: str

    @staticmethod
    def from_json(json_dict: dict) -> GithubRelease:
        assets = [Asset(**asset) for asset in json_dict.get("assets", [])]
        return GithubRelease(
            url=json_dict["url"],
            assets_url=json_dict["assets_url"],
            upload_url=json_dict["upload_url"],
            html_url=json_dict["html_url"],
            id=json_dict["id"],
            author=json_dict["author"],
            node_id=json_dict["node_id"],
            tag_name=json_dict["tag_name"],
            target_commitish=json_dict["target_commitish"],
            name=json_dict["name"],
            draft=json_dict["draft"],
            prerelease=json_dict["prerelease"],
            created_at=json_dict["created_at"],
            published_at=json_dict["published_at"],
            assets=assets,
            tarball_url=json_dict["tarball_url"],
            zipball_url=json_dict["zipball_url"],
            body=json_dict["body"],
        )


@dataclass
class ForkContentsData(AbstractAPIResult):
    id: int
    node_id: str
    name: str
    full_name: str
    private: bool
    owner: UserInfoData
    html_url: str
    description: str | None
    has_discussions: bool
    fork: bool
    url: str
    forks_url: str
    keys_url: str
    collaborators_url: str
    teams_url: str
    hooks_url: str
    issue_events_url: str
    events_url: str
    assignees_url: str
    branches_url: str
    tags_url: str
    blobs_url: str
    git_tags_url: str
    git_refs_url: str
    trees_url: str
    statuses_url: str
    languages_url: str
    stargazers_url: str
    contributors_url: str
    subscribers_url: str
    subscription_url: str
    commits_url: str
    git_commits_url: str
    comments_url: str
    issue_comment_url: str
    contents_url: str
    compare_url: str
    merges_url: str
    archive_url: str
    downloads_url: str
    issues_url: str
    pulls_url: str
    milestones_url: str
    notifications_url: str
    labels_url: str
    releases_url: str
    deployments_url: str
    created_at: str
    updated_at: str
    pushed_at: str
    git_url: str
    ssh_url: str
    clone_url: str
    svn_url: str
    homepage: str | None
    size: int
    stargazers_count: int
    watchers_count: int
    language: str | None
    has_issues: bool
    has_projects: bool
    has_downloads: bool
    has_wiki: bool
    has_pages: bool
    forks_count: int
    mirror_url: str | None
    archived: bool
    disabled: bool
    open_issues_count: int
    license: dict[str, Any] | None
    allow_forking: bool
    is_template: bool
    web_commit_signoff_required: bool
    topics: list[str]
    visibility: str
    forks: int
    open_issues: int
    watchers: int
    default_branch: str


@dataclass
class RepoIndexData(ForkContentsData):
    temp_clone_token: str | None
    custom_properties: dict[str, Any]
    organization: UserInfoData | None
    network_count: int
    subscribers_count: int


@dataclass
class CommitTreeInfo(AbstractAPIResult):
    sha: str
    url: str


@dataclass
class UserInfoData(AbstractAPIResult):
    login: str
    id: int
    node_id: str
    avatar_url: str
    gravatar_id: str
    url: str
    html_url: str
    followers_url: str
    following_url: str
    gists_url: str
    starred_url: str
    subscriptions_url: str
    organizations_url: str
    repos_url: str
    events_url: str
    received_events_url: str
    type: str
    site_admin: bool


@dataclass
class ContributorsInfoData(UserInfoData):
    contributions: int


@dataclass
class AuthorInfoData(AbstractAPIResult):
    name: str
    email: str
    date: str


@dataclass
class CommitDetailData(AbstractAPIResult):
    author: AuthorInfoData
    committer: AuthorInfoData
    message: str
    tree: CommitTreeInfo
    url: str
    comment_count: int
    verification: VerificationInfoData


@dataclass
class VerificationInfoData(AbstractAPIResult):
    verified: bool
    reason: str
    signature: str | None
    payload: str | None


@dataclass
class CommitInfoData(AbstractAPIResult):
    sha: str
    node_id: str
    commit: CommitDetailData
    url: str
    html_url: str
    comments_url: str
    author: UserInfoData
    committer: UserInfoData
    parents: list[ParentCommitInfoData]


@dataclass
class ParentCommitInfoData(CommitTreeInfo):
    html_url: str


@dataclass
class BranchInfoData(AbstractAPIResult):
    name: str
    commit: CommitTreeInfo
    protected: bool


@dataclass
class TreeInfoData(AbstractAPIResult):
    path: str
    mode: str
    type: str
    sha: str
    size: int | None
    url: str


@dataclass
class CompleteRepoData(AbstractAPIResult):
    repo_info: RepoIndexData
    branches: list[BranchInfoData]
    # commits: list[CommitInfoData]
    # issues: list[dict[str, Any]]
    # pulls: list[dict[str, Any]]
    # contributors: list[ContributorsInfoData]
    # releases: list[GithubRelease]
    # tags: list[dict[str, Any]]
    contents: list[ContentInfoData]
    # languages: dict[str, Any]
    forks: list[ForkContentsData]
    # stargazers: list[dict[str, Any]]
    # subscribers: list[dict[str, Any]]
    tree: list[TreeInfoData]

    def __getattr__(self, attr_name: str):
        return self.__dict__[attr_name]

    def get_main_branch_files(self) -> list[ContentInfoData]:
        main_branch = next((branch for branch in self.branches if branch.name == self.repo_info.default_branch), None)
        if not main_branch:
            raise ValueError(f"Main branch {self.repo_info.default_branch} not found in branches")

        return [content for content in self.contents if content.path.startswith(main_branch.commit.sha)]

    @classmethod
    def load_repo(cls, owner: str, repo_name: str, *, timeout: int = 15) -> CompleteRepoData:
        base_url = f"https://api.github.com/repos/{owner}/{repo_name}"

        endpoints = {"repo_info": base_url, "branches": f"{base_url}/branches", "contents": f"{base_url}/contents", "forks": f"{base_url}/forks"}

        repo_data = {}
        import requests

        for key, url in endpoints.items():
            print(f"Fetching {key}...")
            response = requests.get(url, timeout=timeout)
            response.raise_for_status()
            repo_data[key] = response.json()

        # Fetch the tree using the correct default branch
        default_branch = repo_data.get("repo_info", {}).get("default_branch", "main")
        tree_url = f"{base_url}/git/trees/{default_branch}?recursive=1"
        print(f"Fetching tree from {tree_url}...")
        tree_response = requests.get(tree_url, timeout=timeout)
        if tree_response.status_code == 200:
            repo_data["tree"] = tree_response.json().get("tree", [])
        else:
            print(f"Failed to fetch tree from {tree_url}")

        instance = cls(
            repo_info=RepoIndexData.from_dict(repo_data.get("repo_info", {})),
            branches=[BranchInfoData.from_dict(item) for item in repo_data.get("branches", [])],
            contents=[ContentInfoData.from_dict(item) for item in repo_data.get("contents", [])],
            forks=[ForkContentsData.from_dict(item) for item in repo_data.get("forks", [])],
            tree=[TreeInfoData.from_dict(item) for item in repo_data.get("tree", [])],
        )
        print(f"Completed loading of '{base_url}'")
        return instance

    @classmethod
    def load_repo_from_files(cls, folder_path: str) -> CompleteRepoData:
        def load_json(file_path: str) -> Any:
            with open(file_path) as file:
                return json.load(file)

        repo_data = {
            "repo_info": load_json(os.path.join(folder_path, "repo_info.json")),
            "branches": load_json(os.path.join(folder_path, "branches.json")),
            #    "commits": load_json(os.path.join(folder_path, "commits.json")),
            #    "issues": load_json(os.path.join(folder_path, "issues.json")),
            #    "pulls": load_json(os.path.join(folder_path, "pulls.json")),
            #    "contributors": load_json(os.path.join(folder_path, "contributors.json")),
            #    "releases": load_json(os.path.join(folder_path, "releases.json")),
            #    "tags": load_json(os.path.join(folder_path, "tags.json")),
            "contents": load_json(os.path.join(folder_path, "contents.json")),
            #    "languages": load_json(os.path.join(folder_path, "languages.json")),
            "forks": load_json(os.path.join(folder_path, "forks.json")),
            #    "stargazers": load_json(os.path.join(folder_path, "stargazers.json")),
            #    "subscribers": load_json(os.path.join(folder_path, "subscribers.json")),
        }

        return cls(
            repo_info=RepoIndexData.from_dict(repo_data.get("repo_info", {})),
            branches=[BranchInfoData.from_dict(item) for item in repo_data.get("branches", [])],
            #    commits=[CommitInfoData.from_dict(item) for item in repo_data.get("commits", [])],
            #    issues=repo_data.get("issues", []),
            #    pulls=repo_data.get("pulls", []),
            #    contributors=[ContributorsInfoData.from_dict(item) for item in repo_data.get("contributors", [])],
            #    releases=[GithubRelease.from_dict(item) for item in repo_data.get("releases", [])],
            #    tags=repo_data.get("tags", []),
            contents=[ContentInfoData.from_dict(item) for item in repo_data.get("contents", [])],
            #    languages=repo_data.get("languages", {}),
            forks=[ForkContentsData.from_dict(item) for item in repo_data.get("forks", [])],
            #    stargazers=repo_data.get("stargazers", []),
            #    subscribers=repo_data.get("subscribers", []),
        )


def download_github_file(
    url_or_repo: str | tuple[str, str],
    local_path: os.PathLike | str,
    repo_path: os.PathLike | str | None = None,
    timeout: int | None = None,
):
    import requests

    timeout = 180 if timeout is None else timeout
    local_path = Path(local_path).absolute()
    ensure_directory_exists(local_path.parent)

    if isinstance(url_or_repo, tuple):
        owner, repo = url_or_repo
        base_url = f"https://api.github.com/repos/{owner}/{repo}"
    else:
        parsed = urlparse(url_or_repo)
        host = parsed.netloc
        path = parsed.path or ""
        if host == "api.github.com" and path.startswith("/repos/"):
            # Strip trailing path segments to keep only /repos/<owner>/<repo>
            path_parts = path.split("/")
            if len(path_parts) >= 4:
                owner_part, repo_part = path_parts[2], path_parts[3]
                base_url = f"https://api.github.com/repos/{owner_part}/{repo_part}"
            else:
                raise ValueError(f"Invalid GitHub API URL format: {url_or_repo!r}")
        elif host in ("github.com", "raw.githubusercontent.com"):
            parts = [p for p in path.split("/") if p]
            if len(parts) >= 2:
                owner_part, repo_part = parts[0], parts[1]
                base_url = f"https://api.github.com/repos/{owner_part}/{repo_part}"
            else:
                raise ValueError(f"Cannot extract owner/repo from GitHub URL: {url_or_repo!r}")
        else:
            raise ValueError(
                f"URL must be a github.com, raw.githubusercontent.com, or api.github.com URL, got: {url_or_repo!r}"
            )

    if repo_path is not None:
        api_url = f"{base_url}/contents/{PurePath(repo_path).as_posix()}"
        file_info: dict[str, str] = _request_api_data(api_url)
        if file_info["type"] == "file":
            download_url = file_info["download_url"]
        else:
            raise ValueError("The provided repo_path does not point to a file.")
    else:
        download_url = url_or_repo

    with requests.get(download_url, stream=True, timeout=timeout) as r:
        r.raise_for_status()
        with local_path.open("wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)


def download_github_release_asset(
    owner: str,
    repo: str,
    tag_name: str,
    asset_name: str,
    local_path: os.PathLike | str,
    timeout: int | None = None,
) -> bool:
    """Download an asset from a GitHub release.

    Args:
    ----
        owner: GitHub repository owner
        repo: GitHub repository name
        tag_name: Release tag name (e.g., "v1.0.0" or "latest")
        asset_name: Name of the asset to download (e.g., "kits.zip")
        local_path: Local path where the asset will be saved
        timeout: Request timeout in seconds (default: 180)

    Returns:
    -------
        bool: True if download was successful, False otherwise

    Raises:
    ------
        ValueError: If the release or asset is not found
        requests.exceptions.RequestException: If download fails
    """
    import requests

    timeout = 180 if timeout is None else timeout
    local_path = Path(local_path).absolute()
    ensure_directory_exists(local_path.parent)

    # Get release info
    if tag_name.lower() == "latest":
        api_url = f"https://api.github.com/repos/{owner}/{repo}/releases/latest"
    else:
        api_url = f"https://api.github.com/repos/{owner}/{repo}/releases/tags/{tag_name}"

    release_info: dict[str, Any] = _request_api_data(api_url)

    # Find the asset
    assets: list[dict[str, Any]] = release_info.get("assets", [])
    asset_url: str | None = None
    for asset in assets:
        if asset["name"] == asset_name:
            asset_url = asset["browser_download_url"]
            break

    if asset_url is None:
        available_assets = [a["name"] for a in assets]
        raise ValueError(f"Asset '{asset_name}' not found in release '{tag_name}'. Available assets: {', '.join(available_assets) if available_assets else 'none'}")

    # Download the asset
    with requests.get(asset_url, stream=True, timeout=timeout) as r:
        r.raise_for_status()
        with local_path.open("wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)

    return True


def download_github_directory(
    repo: str | tuple[str, str],
    local_dir: os.PathLike | str,
    repo_path: os.PathLike | str,
):
    """This method should not be used due to github's api restrictions. Use download_file to get a .zip of the folder instead."""  # noqa: D404
    repo = PurePath(repo)
    repo_path = PurePath(repo_path)
    api_url = f"https://api.github.com/repos/{repo.as_posix()}/contents/{repo_path.as_posix()}"
    data = _request_api_data(api_url)
    for item in data:
        item_path = Path(item["path"])
        local_path = item_path.relative_to("toolset")

        if item["type"] == "file":
            download_github_file(item["download_url"], Path(local_dir, local_path))
        elif item["type"] == "dir":
            download_github_directory(repo, item_path, local_path)


def download_github_directory_fallback(
    repo: os.PathLike | str,
    local_dir: os.PathLike | str,
    repo_path: os.PathLike | str,
):
    """There were two versions of this function and I can't remember which one worked."""
    repo = PurePath(repo)
    repo_path = PurePath(repo_path)
    api_url = f"https://api.github.com/repos/{repo.as_posix()}/contents/{repo_path.as_posix()}"
    data = _request_api_data(api_url)
    for item in data:
        item_path = Path(item["path"])
        local_path = item_path.relative_to("toolset")

        if item["type"] == "file":
            download_github_file(item["download_url"], local_path)
        elif item["type"] == "dir":
            download_github_directory(repo, item_path, local_path)


@overload
def fetch_repo_index(owner: str, repo: str, branch: str = "master") -> dict[str, str]: ...
@overload
def fetch_repo_index(repo: str, branch: str = "master") -> dict[str, str]: ...
def fetch_repo_index(owner_or_repo: str | tuple[str, str], repo: str = None, branch: str = "master") -> dict[str, str]:
    """Fetches the index of a GitHub repository.

    Args:
    ----
        owner: The owner of the repository.
        repo: The name of the repository.
        branch: The branch of the repository.

    Returns:
    -------
        A dictionary mapping file paths in the repo to their respective GitHub URLs.
    """
    import requests

    if repo is None:
        owner_or_repo, repo = extract_owner_repo(owner_or_repo)
    api_url = f"https://api.github.com/repos/{owner_or_repo}/{repo}/git/trees/{branch}?recursive=1"
    response = requests.get(api_url, timeout=15)
    response.raise_for_status()
    data: dict[str, Any] = response.json()

    repo_index = {item["path"]: f"https://github.com/{owner_or_repo}/{repo}/blob/{branch}/{item['path']}" for item in data.get("tree", []) if item["type"] == "blob"}
    return repo_index


def _request_api_data(api_url: str) -> Any:
    import requests

    response: requests.Response = requests.get(api_url, timeout=15)
    response.raise_for_status()
    return response.json()


@overload
def get_api_url(owner: str, repo: str) -> str: ...
@overload
def get_api_url(repo: str) -> str: ...
def get_api_url(owner: str, repo: str = None) -> str:
    if repo is None:
        owner, repo = extract_owner_repo(owner)
    return f"https://api.github.com/repos/{owner}/{repo}"


@overload
def get_forks_url(owner: str, repo: str) -> str: ...
@overload
def get_forks_url(repo: str) -> str: ...
def get_forks_url(owner: str, repo: str = None) -> str:
    if repo is None:
        owner, repo = extract_owner_repo(owner)
    return f"https://api.github.com/repos/{owner}/{repo}/forks"


@overload
def get_main_url(owner: str, repo: str) -> str: ...
@overload
def get_main_url(repo: str) -> str: ...
def get_main_url(owner: str, repo: str = None) -> str:
    if repo is None:
        owner, repo = extract_owner_repo(owner)
    return f"https://github.com/{owner}/{repo}"


def extract_owner_repo_from_api_url(url: str) -> tuple[str, str]:
    pattern = re.compile(r"https://api\.github\.com/repos/([^/]+)/([^/]+)")
    match = pattern.match(url)
    if not match:
        raise ValueError("Invalid GitHub API URL")
    return match[1], match[2]


def extract_owner_repo_from_raw_url(url: str) -> tuple[str, str]:
    pattern = re.compile(r"https://raw\.githubusercontent\.com/([^/]+)/([^/]+)")
    match = pattern.match(url)
    if not match:
        raise ValueError("Invalid GitHub raw content URL")
    return match[1], match[2]


def extract_owner_repo_from_main_url(url: str) -> tuple[str, str]:
    pattern = re.compile(r"https://github\.com/([^/]+)/([^/]+)")
    match = pattern.match(url)
    if not match:
        raise ValueError("Invalid GitHub main URL")
    return match[1], match[2]


def extract_owner_repo(url: str) -> tuple[str, str]:
    parsed = urlparse(url)
    host = parsed.netloc
    path = parsed.path or ""

    # Normalize back to a canonical URL for the existing regex helpers.
    normalized_url = f"https://{host}{path}"

    if host == "api.github.com" and path.startswith("/repos/"):
        return extract_owner_repo_from_api_url(normalized_url)
    if host == "raw.githubusercontent.com":
        return extract_owner_repo_from_raw_url(normalized_url)
    if host == "github.com":
        return extract_owner_repo_from_main_url(normalized_url)
    raise ValueError(f"Invalid GitHub URL format: {url}")


if __name__ == "__main__":
    import sys

    from toolset.__main__ import onAppCrash

    sys.excepthook = onAppCrash
    test1 = CompleteRepoData.load_repo_from_files(r"C:\GitHub\PyKotor\KOTORCommunityPatches_Vanilla_KOTOR_Script_Source\json files")
    test1_dict = test1.to_dict()
    print(json.dumps(test1_dict, indent=4))



"""Render BWM walkmesh to validation PNGs using matplotlib 3D.

Used by the CLI when --render-png is passed to walkmesh-rebuild. Importing this
module requires matplotlib; the CLI catches ImportError and instructs the user
to install pykotor[render].
"""

from __future__ import annotations

import math
import pathlib

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.figure import Figure
from matplotlib.lines import Line2D
from matplotlib.patches import Patch
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Line3DCollection, Poly3DCollection

from pykotor.resource.formats.bwm.bwm_data import BWM

# Palette (RGBA)
WALKABLE_COLOR = (0.1, 0.2, 0.5, 0.6)
UNWALKABLE_COLOR = (0.2, 0.2, 0.2, 0.6)
ARROW_COLOR = (1.0, 0.3, 0.0, 1.0)
GRID_COLOR = (0.5, 0.5, 0.5, 0.4)
EDGE_COLOR = (0.1, 0.1, 0.1, 1.0)
NORMAL_COLOR = (0.0, 0.8, 0.8, 0.8)


def render_bwm_to_pngs(
    bwm: BWM,
    output_stem: pathlib.Path,
    *,
    dpi: int = 300,
    size_inches: tuple[float, float] = (6.4, 3.6),
) -> list[pathlib.Path]:
    """Render the walkmesh to 3–4 validation PNGs (views) next to output_stem.

    Writes {output_stem}_view1.png … _view4.png. No logging; caller logs paths.

    Args:
    ----
        bwm: Rebuilt BWM in memory.
        output_stem: Directory + stem (e.g. output_dir / "203tell_rebuild").
        dpi: DPI for saved figures.
        size_inches: Figure size (width, height) in inches.

    Returns:
    -------
        List of paths to the written PNG files.
    """
    verts_list = [(v.x, v.y, v.z) for v in bwm.vertices()]
    walkable = bwm.walkable_faces()
    unwalkable = bwm.unwalkable_faces()
    faces = walkable + unwalkable
    bbmin, bbmax = bwm.box()
    edges = list(bwm.edges())
    n_perim = len(edges)
    trans_edges = [(i, e) for i, e in enumerate(edges) if e.transition >= 0]
    n_trans = len(trans_edges)
    n_verts = len(verts_list)
    n_walk = len(walkable)
    n_unwalk = len(unwalkable)
    extent_x = bbmax.x - bbmin.x
    extent_y = bbmax.y - bbmin.y
    extent_z = bbmax.z - bbmin.z
    diag = math.sqrt(extent_x**2 + extent_y**2 + extent_z**2)
    arrow_scale = max(0.25, 0.07 * diag)
    gizmo_scale = 0.05 * diag
    normal_scale = 0.08 * diag

    # Build triangle verts and face colors (identity-based vertex lookup)
    verts_by_id = {id(v): i for i, v in enumerate(bwm.vertices())}
    tri_verts: list[np.ndarray] = []
    face_colors: list[tuple[float, float, float, float]] = []
    for face in faces:
        i1 = verts_by_id[id(face.v1)]
        i2 = verts_by_id[id(face.v2)]
        i3 = verts_by_id[id(face.v3)]
        tri_verts.append(
            np.array([
                verts_list[i1],
                verts_list[i2],
                verts_list[i3],
            ]),
        )
        face_colors.append(
            WALKABLE_COLOR if face.material.walkable() else UNWALKABLE_COLOR,
        )

    # Arrows and transition labels: perimeter edges with transition >= 0
    arrow_segments: list[tuple[tuple[float, float, float], tuple[float, float, float]]] = []
    transition_labels: list[tuple[float, float, float, int]] = []  # (x, y, z, trans_id)
    for trans_num, (_, edge) in enumerate(trans_edges, start=1):
        mid, direction = BWM.edge_inward_direction_xy(edge.face, edge.index)
        end = (
            mid.x + arrow_scale * direction.x,
            mid.y + arrow_scale * direction.y,
            mid.z,
        )
        arrow_segments.append(((mid.x, mid.y, mid.z), end))
        transition_labels.append((mid.x, mid.y, mid.z, trans_num))

    # Grid at Z = bbmin.z, XY
    gx = np.linspace(bbmin.x, bbmax.x, max(2, int((bbmax.x - bbmin.x) / 1.0) + 1))
    gy = np.linspace(bbmin.y, bbmax.y, max(2, int((bbmax.y - bbmin.y) / 1.0) + 1))
    z_grid = bbmin.z
    grid_lines: list[tuple[tuple[float, float, float], tuple[float, float, float]]] = []
    for x in gx:
        grid_lines.append(((x, gy[0], z_grid), (x, gy[-1], z_grid)))
    for y in gy:
        grid_lines.append(((gx[0], y, z_grid), (gx[-1], y, z_grid)))

    # Axis gizmo at corner (bbmin)
    gizmo_lines = [
        ((bbmin.x, bbmin.y, bbmin.z), (bbmin.x + gizmo_scale, bbmin.y, bbmin.z)),  # X red
        ((bbmin.x, bbmin.y, bbmin.z), (bbmin.x, bbmin.y + gizmo_scale, bbmin.z)),  # Y green
        ((bbmin.x, bbmin.y, bbmin.z), (bbmin.x, bbmin.y, bbmin.z + gizmo_scale)),  # Z blue
    ]

    # Face normals (every 10th face)
    normal_segments: list[tuple[tuple[float, float, float], tuple[float, float, float]]] = []
    for i, face in enumerate(faces):
        if i % 10 != 0:
            continue
        c = face.centre()
        n = face.normal()
        normal_segments.append(
            ((c.x, c.y, c.z), (c.x + normal_scale * n.x, c.y + normal_scale * n.y, c.z + normal_scale * n.z)),
        )

    extent = (
        bbmax.x - bbmin.x,
        bbmax.y - bbmin.y,
        bbmax.z - bbmin.z,
    )
    aspect = (max(extent[0], 1e-6), max(extent[1], 1e-6), max(extent[2], 1e-6))

    views = [
        ("view1", 90, 0, True, True),   # top-down
        ("view2", 35, 30, True, True),  # perspective 1
        ("view3", 35, 210, True, True), # perspective 2
        ("view4", 35, 30, False, True), # wireframe emphasis (edges only over solid)
    ]

    title_parts = [
        f"V:{n_verts}  F:{n_walk}w/{n_unwalk}u  Perimeter:{n_perim}  Transitions:{n_trans}",
        f"Bbox: X[{bbmin.x:.1f},{bbmax.x:.1f}] Y[{bbmin.y:.1f},{bbmax.y:.1f}] Z[{bbmin.z:.1f},{bbmax.z:.1f}]  Extent: {extent_x:.1f} x {extent_y:.1f} x {extent_z:.1f}",
    ]
    suptitle: str = "  |  ".join(title_parts)

    written: list[pathlib.Path] = []
    for view_name, elev, azim, draw_faces, draw_edges in views:
        fig: Figure = plt.figure(figsize=size_inches)
        fig.suptitle(suptitle, fontsize=8, wrap=True)
        ax = fig.add_subplot(111, projection="3d")
        # set_box_aspect can raise "not enough values to unpack (expected 5, got 0)" on some
        # matplotlib 3D backends when axis bounds are uninitialized; skip it so the plot still renders.
        try:
            ax.set_box_aspect(aspect)
        except (ValueError, TypeError):
            pass
        ax.view_init(elev=elev, azim=azim)

        if draw_faces:
            coll = Poly3DCollection(tri_verts, facecolors=face_colors, alpha=0.6, edgecolors="none")
            ax.add_collection3d(coll)
        if draw_edges:
            edge_verts = [t for t in tri_verts]
            edge_coll = Poly3DCollection(edge_verts, facecolors="none", edgecolors=EDGE_COLOR, linewidths=0.8)
            ax.add_collection3d(edge_coll)
        if grid_lines:
            lc = Line3DCollection(grid_lines, colors=GRID_COLOR, linewidths=0.5)
            ax.add_collection3d(lc)
        if arrow_segments:
            ac = Line3DCollection(arrow_segments, colors=ARROW_COLOR, linewidths=1.2)
            ax.add_collection3d(ac)
        if gizmo_lines:
            ax.plot(
                [gizmo_lines[0][0][0], gizmo_lines[0][1][0]],
                [gizmo_lines[0][0][1], gizmo_lines[0][1][1]],
                [gizmo_lines[0][0][2], gizmo_lines[0][1][2]],
                color="red",
                linewidth=2,
            )
            ax.plot(
                [gizmo_lines[1][0][0], gizmo_lines[1][1][0]],
                [gizmo_lines[1][0][1], gizmo_lines[1][1][1]],
                [gizmo_lines[1][0][2], gizmo_lines[1][1][2]],
                color="green",
                linewidth=2,
            )
            ax.plot(
                [gizmo_lines[2][0][0], gizmo_lines[2][1][0]],
                [gizmo_lines[2][0][1], gizmo_lines[2][1][1]],
                [gizmo_lines[2][0][2], gizmo_lines[2][1][2]],
                color="blue",
                linewidth=2,
            )
        if normal_segments:
            nc = Line3DCollection(normal_segments, colors=NORMAL_COLOR, linewidths=0.5)
            ax.add_collection3d(nc)
        for x, y, z, tid in transition_labels:
            ax.text(x, y, z, f" {tid} ", fontsize=7, color=ARROW_COLOR[:3], ha="center")

        ax.set_xlim(bbmin.x, bbmax.x)
        ax.set_ylim(bbmin.y, bbmax.y)
        ax.set_zlim(bbmin.z, bbmax.z)
        ax.set_xlabel("X (world)")
        ax.set_ylabel("Y (world)")
        ax.set_zlabel("Z (world)")
        # In-figure legend: walkable, unwalkable, perimeter arrows (ID = transition/door index)
        legend_elements = [
            Patch(facecolor=WALKABLE_COLOR, edgecolor="none", label="Walkable"),
            Patch(facecolor=UNWALKABLE_COLOR, edgecolor="none", label="Unwalkable"),
            Line2D([0], [0], color=ARROW_COLOR, linewidth=2, label="Transition (door) arrow"),
        ]
        ax.legend(handles=legend_elements, loc="upper left", fontsize=6)

        path = output_stem.parent / f"{output_stem.name}_{view_name}.png"
        fig.savefig(path, dpi=dpi, bbox_inches="tight")
        plt.close(fig)
        written.append(path)

    return written



from __future__ import annotations

from typing import TYPE_CHECKING

from qtpy.QtWidgets import QMainWindow, QTreeWidget, QTreeWidgetItem

from utility.gui.qt.widgets.itemviews.treeview import RobustTreeView

if TYPE_CHECKING:
    from qtpy.QtWidgets import QMenu, QWidget


class RobustTreeWidget(QTreeWidget, RobustTreeView):
    """A tree widget that supports common features and settings."""

    def __new__(cls, *args, **kwargs):
        # For PySide6 compatibility with multiple inheritance
        return QTreeWidget.__new__(cls)

    def __init__(
        self,
        parent: QWidget | None = None,
        *args,
        use_columns: bool = False,
        should_call_qt_init: bool = True,
        **kwargs,
    ):
        if should_call_qt_init:
            QTreeWidget.__init__(self, parent, *args, **kwargs)

        RobustTreeView.__init__(
            self,
            parent,
            *args,
            use_columns=use_columns,
            should_call_qt_init=False,  # QTreeView.__init__ should not be called since we already call QTreeWidget.__init__ and handle that argument.
            **kwargs,
        )

    def build_context_menu(self, parent: QWidget | None = None) -> QMenu:
        menu = super().build_context_menu(parent)

        tree_widget_menu = menu.addMenu("TreeWidget")
        advanced_menu = tree_widget_menu.addMenu("Advanced")

        # Tree-specific actions
        self._add_menu_action(
            tree_widget_menu,
            "Expand Current Item",
            lambda: self.isExpanded(self.currentIndex()),
            lambda x: self.expandItem(self.currentItem()) if x else self.collapseItem(self.currentItem()),
            "expandCurrentItem",
        )
        self._add_simple_action(tree_widget_menu, "Scroll to Current Item", lambda: self.scrollToItem(self.currentItem()))
        self._add_menu_action(
            tree_widget_menu,
            "Sort Items",
            self.isSortingEnabled,
            self.setSortingEnabled,
            "sortingEnabled",
        )
        self._add_menu_action(
            tree_widget_menu,
            "Column Count",
            self.columnCount,
            self.setColumnCount,
            "columnCount",
            param_type=int,
        )

        # Advanced actions
        self._add_menu_action(
            advanced_menu,
            "Header Hidden",
            self.isHeaderHidden,
            self.setHeaderHidden,
            "headerHidden",
        )
        self._add_menu_action(
            advanced_menu,
            "Root Is Decorated",
            self.rootIsDecorated,
            self.setRootIsDecorated,
            "rootIsDecorated",
        )
        self._add_menu_action(
            advanced_menu,
            "Items Expandable",
            self.itemsExpandable,
            self.setItemsExpandable,
            "itemsExpandable",
        )
        self._add_menu_action(
            advanced_menu,
            "Animated",
            self.isAnimated,
            self.setAnimated,
            "animated",
        )
        self._add_menu_action(
            advanced_menu,
            "Allow Expand on Single Click",
            self.allColumnsShowFocus,
            self.setAllColumnsShowFocus,
            "allColumnsShowFocus",
        )
        self._add_menu_action(
            advanced_menu,
            "Word Wrap",
            self.wordWrap,
            self.setWordWrap,
            "wordWrap",
        )

        # Item manipulation actions
        item_menu = tree_widget_menu.addMenu("Item Actions")
        self._add_simple_action(item_menu, "Add Top Level Item", lambda: self.addTopLevelItem(QTreeWidgetItem(["New Item"])))
        self._add_simple_action(item_menu, "Remove Current Item", lambda: self.takeTopLevelItem(self.indexOfTopLevelItem(self.currentItem())) if self.currentItem() else None)
        self._add_simple_action(item_menu, "Clear All Items", self.clear)
        self._add_simple_action(item_menu, "Edit Current Item", lambda: self.editItem(self.currentItem(), 0) if self.currentItem() else None)

        return menu


if __name__ == "__main__":
    import sys

    from qtpy.QtWidgets import QApplication, QPushButton, QVBoxLayout, QWidget

    app = QApplication(sys.argv)

    # Create main window
    window = QMainWindow()
    central_widget = QWidget()
    window.setCentralWidget(central_widget)
    layout = QVBoxLayout(central_widget)

    # Create RobustTreeWidget
    tree_widget = RobustTreeWidget()
    layout.addWidget(tree_widget)

    # Add some sample items
    for i in range(5):
        parent = QTreeWidgetItem(tree_widget, [f"Parent {i + 1}"])
        for j in range(3):
            QTreeWidgetItem(parent, [f"Child {j + 1} of Parent {i + 1}"])

    # Add a button to show context menu
    def show_context_menu():
        menu = tree_widget.build_context_menu()
        menu.exec(tree_widget.mapToGlobal(tree_widget.rect().center()))

    context_menu_button = QPushButton("Show Context Menu")
    context_menu_button.clicked.connect(show_context_menu)
    layout.addWidget(context_menu_button)

    window.setGeometry(100, 100, 400, 500)
    window.setWindowTitle("RobustTreeWidget Test")
    window.show()

    sys.exit(app.exec())



from __future__ import annotations

import re

from qtpy.QtCore import QRegularExpression
from qtpy.QtGui import QRegularExpressionValidator


class FileNameValidator(QRegularExpressionValidator):
    INVALID_CHARACTERS = '<>:"/\\|?*'

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setRegularExpression(self.get_regular_expression())

    @classmethod
    def get_regular_expression(cls) -> QRegularExpression:
        # Create a regular expression pattern that matches valid filename characters
        pattern = f"^[^{re.escape(cls.INVALID_CHARACTERS)}]*$"
        return QRegularExpression(pattern)

    def regularExpression(self) -> QRegularExpression:
        return super().regularExpression()



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any

from qtpy.QtCore import Qt
from qtpy.QtGui import QFont, QIcon
from qtpy.QtWidgets import (
    QCheckBox,
    QDialog,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QScrollArea,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)


@dataclass(frozen=True)
class FileProperties:
    name: str
    path: str
    type: str
    size: str
    size_on_disk: str
    created: str
    modified: str
    accessed: str
    mime_type: str
    owner: str
    group: str
    permissions: str
    inode: int
    num_hard_links: int
    device: int
    is_symlink: bool
    symlink_target: str
    md5: str
    sha1: str
    sha256: str
    is_hidden: bool
    is_system: bool
    is_archive: bool
    is_compressed: bool
    is_encrypted: bool
    is_readonly: bool
    is_temporary: bool
    extension: str


class FilePropertiesDialog(QDialog):
    def __init__(self, properties: FileProperties, parent: QWidget | None = None):
        super().__init__(parent)
        self.prop = properties
        self.setWindowTitle(f"Properties - {self.prop.name}")
        self.setMinimumSize(500, 600)
        self.setup_ui()

    def setup_ui(self):
        main_layout = QVBoxLayout(self)

        # File icon and name at the top
        top_layout = QHBoxLayout()
        icon_label = QLabel()
        icon_label.setPixmap(QIcon.fromTheme(self.prop.type.lower()).pixmap(64, 64))
        top_layout.addWidget(icon_label)
        name_label = QLabel(self.prop.name)
        name_label.setFont(QFont("Arial", 16, QFont.Bold))
        top_layout.addWidget(name_label)
        top_layout.addStretch()
        main_layout.addLayout(top_layout)

        # Tabs for different property categories
        tab_widget = QTabWidget()
        tab_widget.addTab(self.create_general_tab(), "General")
        tab_widget.addTab(self.create_security_tab(), "Security")
        tab_widget.addTab(self.create_details_tab(), "Details")
        main_layout.addWidget(tab_widget)

        # OK button at the bottom
        ok_button = QPushButton("OK")
        ok_button.clicked.connect(lambda: self.accept())
        main_layout.addWidget(ok_button, alignment=Qt.AlignRight)

    def _get_relative_time(self, timestamp: datetime) -> str:
        now = datetime.now().astimezone()
        timestamp = timestamp.replace(tzinfo=now.tzinfo)  # Make timestamp offset-aware
        delta = now - timestamp
        if delta.days > 365:  # noqa: PLR2004
            return f"{delta.days // 365} years ago"
        if delta.days > 30:  # noqa: PLR2004
            return f"{delta.days // 30} months ago"
        if delta.days > 0:
            return f"{delta.days} days ago"
        if delta.seconds > 3600:  # noqa: PLR2004
            return f"{delta.seconds // 3600} hours ago"
        if delta.seconds > 60:  # noqa: PLR2004
            return f"{delta.seconds // 60} minutes ago"
        return "Just now"

    def create_general_tab(self):
        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)
        content_widget = QWidget()
        layout = QVBoxLayout(content_widget)

        # Type and Location
        type_group = self.create_group_box(
            "Type and Location",
            [
                ("Type:", self.prop.type),
                ("Location:", self.prop.path),
                ("Size:", f"{self.prop.size} ({self.prop.size_on_disk} on disk)"),
            ],
        )
        layout.addWidget(type_group)

        # Dates
        dates_group = self.create_group_box(
            "Dates",
            [
                ("Created:", self.prop.created),
                ("Modified:", self.prop.modified),
                ("Accessed:", self.prop.accessed),
            ],
        )
        layout.addWidget(dates_group)

        # Attributes
        attributes = [
            ("Hidden", self.prop.is_hidden),
            ("Read-only", self.prop.is_readonly),
            ("System", self.prop.is_system),
            ("Archive", self.prop.is_archive),
            ("Temporary", self.prop.is_temporary),
        ]
        attributes_group = self.create_group_box("Attributes", attributes, is_checkbox=True)
        layout.addWidget(attributes_group)

        scroll_area.setWidget(content_widget)
        return scroll_area

    def create_security_tab(self):
        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)
        content_widget = QWidget()
        layout = QVBoxLayout(content_widget)

        # Owner and Permissions
        security_group = self.create_group_box(
            "Security",
            [
                ("Owner:", self.prop.owner),
                ("Group:", self.prop.group),
                ("Permissions:", self.prop.permissions),
            ],
        )
        layout.addWidget(security_group)
        layout.addStretch()

        scroll_area.setWidget(content_widget)
        return scroll_area

    def create_details_tab(self):
        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)
        content_widget = QWidget()
        layout = QVBoxLayout(content_widget)

        # File Details
        details_group = self.create_group_box(
            "File Details",
            [
                ("MIME Type:", self.prop.mime_type),
                ("Extension:", self.prop.extension),
                ("Is Symlink:", "Yes" if self.prop.is_symlink else "No"),
                ("Symlink Target:", self.prop.symlink_target if self.prop.is_symlink else "N/A"),
                ("Inode:", str(self.prop.inode)),
                ("Number of Hard Links:", str(self.prop.num_hard_links)),
                ("Device:", str(self.prop.device)),
            ],
        )
        layout.addWidget(details_group)

        # File Hashes
        hashes_group = self.create_group_box(
            "File Hashes",
            [
                ("MD5:", self.prop.md5),
                ("SHA1:", self.prop.sha1),
                ("SHA256:", self.prop.sha256),
            ],
        )
        layout.addWidget(hashes_group)

        scroll_area.setWidget(content_widget)
        return scroll_area

    def create_group_box(
        self,
        title: str,
        items: list[tuple[str, Any]],
        *,
        is_checkbox: bool = False,
    ) -> QGroupBox:
        group = QGroupBox(title)
        layout = QVBoxLayout()
        for label, value in items:
            if is_checkbox:
                checkbox = QCheckBox(label)
                checkbox.setChecked(value)
                checkbox.setEnabled(False)
                layout.addWidget(checkbox)
            else:
                item_layout = QHBoxLayout()
                label_widget = QLabel(label)
                label_widget.setFont(QFont("Roboto", 9, QFont.Bold))
                item_layout.addWidget(label_widget)
                value_widget = QLabel(value)
                value_widget.setTextInteractionFlags(Qt.TextSelectableByMouse)
                item_layout.addWidget(value_widget, 1)
                layout.addLayout(item_layout)
        group.setLayout(layout)
        return group



from __future__ import annotations

import threading
import tkinter as tk

from tkinter import ttk
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Literal


class LoadingDialog(tk.Toplevel):
    def __init__(
        self,
        parent: tk.Tk,
        title: str = "Loading",
        bounce_speed: int | Literal["idle"] | None = None,
        pb_length: int = 200,
        func: Callable | None = None,
    ):
        super().__init__(parent)

        # Set basic window properties
        self.title(title)
        self.geometry("300x100")
        self.transient(parent)
        self.grab_set()

        # Message label
        self.msg_lbl: tk.Label = tk.Label(self, text="Please wait...")
        self.msg_lbl.pack(padx=10, pady=5)

        # Progress bar
        self.load_bar: ttk.Progressbar = ttk.Progressbar(self, mode="indeterminate", maximum=100, value=0, length=pb_length)
        self.load_bar.pack(padx=10, pady=(0, 10))
        self.load_bar.start(bounce_speed)

        # Function to execute
        self.func: Callable | None = func

        # Start work in a new thread
        self.start_work_thread()

    def start_work_thread(self):
        if self.func is not None:
            self.work_thread = threading.Thread(target=self.work_task, name="LoadingDialog_worker_thread")
            self.work_thread.start()

    def work_task(self):
        if self.func is not None:
            self.func()
        # Stop progress bar and close dialog after work is done
        self.load_bar.stop()
        self.destroy()



from __future__ import annotations

import shlex
import subprocess
import sys

from typing import Iterable, Mapping


def run_subprocess(
    program: str,
    args: Iterable[str],
    *,
    use_shell_execute: bool,
    hide_process: bool,
    timeout: int | None = None,
    creationflags: int = 0,
    env: Mapping[str, str] | None = None,
    check: bool = True,
    encoding: str = "utf-8",
    stdout_option: int | None = None,
    stderr_option: int | None = None,
) -> tuple[int, bytes, bytes]:
    stdout_option = subprocess.PIPE if hide_process else subprocess.DEVNULL
    stderr_option = subprocess.PIPE if hide_process else subprocess.DEVNULL
    escaped_args = [shlex.quote(arg) for arg in args]
    result: subprocess.CompletedProcess[bytes] = subprocess.run(  # noqa: S603
        [program, *escaped_args],
        shell=use_shell_execute,  # noqa: S603
        stdout=stdout_option,
        stderr=stderr_option,
        check=check,
        timeout=timeout,
        creationflags=creationflags,
        env=env,
    )
    return result.returncode, result.stdout, result.stderr


def is_frozen() -> bool:
    # Check for sys attributes - legitimate use of getattr for optional runtime attributes
    return getattr(sys, "frozen", False) or getattr(sys, "_MEIPASS", False)
